<?php
    require_once KROOT . '/classes/KJsonRpcClient.php';

    /**
     * Portability manager import options
     */
    class KGatewayPortabilityOptions
    {
        public $fieldDelimiter = ',';
        public $fieldClosure = '"';
        public $ignoreFirstLine = false;
        public $incremental = false;
        public $ignoreErrors = false;
        public $removeAllContentFirst = false;
    }

    /**
     * Portability manager import helper
     */
    class KGatewayPortabilityImport
    {
        static private $me = NULL;

        private $temporaryDirectory = '/dev/shm';
        private $kgatewaymanager = '/usr/sbin/kgateway-portability-manager';
        private $gunzip = '/bin/gunzip';
        private $rm = '/bin/rm';
        private $rpc = null;
        private $driver = null;

        const DRIVER_DISABLED = 'DISABLED';
        const DRIVER_LOCAL_DATABASE = 'LOCAL-DATABASE';
        const DRIVER_WEBSERVICE = 'WEBSERVICE';

        private function __construct( )
        {
            $this->rpc = new KJsonRpcClient( 'kgateway' );
            try
            {
                $this->driver = $this->rpc->PortabilityDriver( );
            }
            catch( Exception $e )
            {
                $this->driver = self::DRIVER_DISABLED;
            }
        }

        static public function getInstance( )
        {
            if( is_null( self::$me ) )
            {
                self::$me = new self( );
            }
            return self::$me;
        }

        /**
         * Available input file types
         *
         * @return array
         */
        public function getAvailableImportTypes( )
        {
            switch( $this->driver )
            {
            case self::DRIVER_LOCAL_DATABASE:
                return array( 'import-service-providers', 'import-numbers', 'import-prefixes' );
            case self::DRIVER_WEBSERVICE:
                return array( 'import-service-providers' );
            }
            return array( );
        }

        /**
         * Available input file field formats
         *
         * @return array
         */
        public function getAvailableFormats( )
        {
            switch( $this->driver )
            {
            case self::DRIVER_LOCAL_DATABASE:
                return array( 'IGNORE', 'NUMBER', 'CODE', 'SERVICEPROVIDER', 'OPERATION' );
            case self::DRIVER_WEBSERVICE:
                return array( 'IGNORE', 'CODE', 'SERVICEPROVIDER', 'OPERATION' );
            }
            return array( );
        }

        /**
         * Create temporary file name
         *
         * @param string $append Append this string to the end of temporary file
         */
        public function createTemporaryFileName( $append )
        {
            return $this->temporaryDirectory . '/' . str_replace( '.', '_', microtime( true ) ) . '_' . $append;
        }

        /**
         * Create portability manager call suitable for shell
         *
         * @param array $opts Options to be supplied to the manager
         */
        private function createManagerCall( array $opts )
        {
            $_opts = join( ' ', $opts );
            return "{$this->kgatewaymanager} {$_opts}";
        }

        /**
         * Run portability manager in background to import file into database
         *
         * @param string $filename TEMPORARY file to be imported (will be removed
         *                         when done)
         * @param string $type Import type
         * @param string $format input file format
         * @param KGatewayPortabilityOptions $options Import options
         */
        public function run( $filename, $type, $format, KGatewayPortabilityOptions $options )
        {
            $opts = array( );
            $opts[] = escapeshellarg( $type );
            $opts[] = escapeshellarg( $format );
            if( $options->ignoreErrors )
            {
                $opts[] = '-I';
            }
            if( $options->ignoreFirstLine )
            {
                $opts[] = '-1';
            }
            if( $options->incremental )
            {
                $opts[] = '-i';
            }
            if( $options->fieldDelimiter )
            {
                $opts[] = "-d=\\{$options->fieldDelimiter}";
            }
            if( $options->fieldClosure )
            {
                $opts[] = "-c=\\{$options->fieldClosure}";
            }
            if( $options->removeAllContentFirst )
            {
                $opts[] = '-r';
            }

            if( substr( $filename, strlen( $filename ) - 2 ) == "gz" )
            {
                $call = "{$this->gunzip} -c {$filename} | " . $this->createManagerCall( $opts );
            }
            else
            {
                $opts[] = "-F={$filename}";
                $call = $this->createManagerCall( $opts );
            }

            // always remove the temporary file when done
            $call = "( {$call} ; {$this->rm} -f {$filename} ; ) > /dev/null 2>&1 &";
            //$call = "( {$this->rm} -f {$filename} ; ) > /dev/null 2>&1 &";

            exec( $call );
        }

    }
