<?php

    class KGatewayNetworkException extends Exception
    {
    }

    class KGatewayNetwork
    {
        /**
         * array Network interfaces
         */
        protected $interfaces = array( );
        /**
         * int Input/Output error
         */
        const EIO = 1;
        /**
         * int Runtime error
         */
        const ERUNTIME = 2;
        /**
         * int Invalid parameter/value
         */
        const EINVALID = 3;
        /**
         * int Permission denied
         */
        const ENOPERM = 4;
        /**
         * array Supported methods
         */
        static public $methods = array( 'inet' => array( 'manual', 'dhcp', 'static' ) );
        // @formatter:off
        /**
         * array Stanzas from interface(5)
         *
         * Format: __family__ => array
         *           __method__ => array
         *             valid stanza => array
         *               verification method => string|array
         *
         * The verification method can be a string or array. All verification
         * methods accept at least one value (the value to be verified) and
         * return true or false. Optionaly the methods can receive a second
         * value that should be supplied defining this verification method
         * as array and the second position should be the value to be supplied
         * as second parameter to the 'real' verification method.
         */
        static protected $stanzas = array(
            'inet' => array(
                'loopback' => array( ),
                'static' => array(
                    'address' => array( 'checkIpv4', array( 'check-ebs-subnet' => true, 'restrict-first-and-last' => true ) ),
                    'netmask' => 'checkIpv4',
                    'broadcast' => array( 'checkIpv4', array( 'empty' => true ) ),
                    'network' => array( 'checkIpv4', array( 'empty' => true ) ),
                    'metric' => array( 'checkInteger', array( 'empty' => true ) ),
                    'gateway' => array( 'checkIpv4', array( 'check-ebs-subnet' => true, 'empty' => true, 'restrict-first-and-last' => true ) ),
                    // issue #15497
                    'x-khomp-gateway' => array( 'checkIpv4', array( 'check-ebs-subnet' => true, 'empty' => true, 'restrict-first-and-last' => true ) ),
                    'pointopoint' => array( 'checkIpv4', true ),
                    'media' => null,
                    'hwaddress' => array( 'checkMacAddress', array( 'empty' => true ) ),
                    'mtu' => array( 'checkInteger', array( 'min' => 64, 'max' => 1500, 'empty' => true ) )
                ),
                'dhcp' => array(
                    'hostname' => 'checkFqdn',
                    'leasehours' => 'checkInteger',
                    'leasetime' => 'checkInteger',
                    'vendor' => null,
                    'client' => 'checkAlwaysInvalid',
                    'hwaddress' => 'checkMacAddress'
                ),
                'bootp' => array(
                    'bootfile' => null,
                    'server' => 'checkIpv4OrIpv6',
                    'hwaddr' => 'checkMacAddress'
                ),
                'ppp' => array(
                    'provider' => null
                ),
                'wvdial' => array(
                    'provider' => null
                ),
                'manual' => array( )
            ),
            'ipx' => array(
                'static' => array(
                    'frame' => 'checkIpxFrame',
                    'netnum' => 'checkIpxNetNum'
                ),
                'dynamic' => array(
                    'frame' => 'checkIpxFrame'
                )
            ),
            'inet6' => array(
                'loopback' => array( ),
                'static' => array(
                    'address' => 'checkIpv6',
                    'netmask' => 'checkIpv6Mask',
                    'gateway' => array( 'checkIpv6', array( 'empty' => true ) ),
                    // issue #15497
                    'x-khomp-gateway' => array( 'checkIpv6', array( 'empty' => true ) ),
                    'media' => null,
                    'hwaddress' => array( 'checkMacAddress', array( 'empty' => true ) ),
                    'mtu' => array( 'checkInteger', array( 'min' => 64, 'max' => 1500, 'empty' => true ) )
                ),
                'auto' => array( ),
                'manual' => array( )
            )
        );
        /**
         * array Stanzas from resolvconf(8)
         *
         * Format: stanza => array
         *           verification method => string|array
         */
        static protected $stanzasResolvconf = array(
            'dns-nameservers' => array( 'checkIpv4OrIpv6List', array( 'empty' => true, 'restrict-first-and-last' => true ) ),
            'dns-search' => array( 'checkFqdn', array( 'empty' => true ) )
        );
        // @formatter:on
        /**
         * array Supported dns methods
         */
        static public $dnsmethods = array( 'dhcp', 'static' );
        /**
         * string File name
         */
        const INTERFACES_FILE = '/etc/network/interfaces';

        static private $self = null;

        private function __construct( )
        {
            /* VOID */
        }

        /**
         * Get singleton instance
         */
        static public function getInstance( )
        {
            if( is_null( self::$self ) )
            {
                self::$self = new KGatewayNetwork( );
                self::$self->load( );
            }
            return self::$self;
        }

        /**
         * Parse the file and create internal representation
         */
        public function load( )
        {
            if( !file_exists( self::INTERFACES_FILE ) )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-configuration-not-found', self::EIO );
            }
            $lines = file( self::INTERFACES_FILE, FILE_IGNORE_NEW_LINES );
            if( empty( $lines ) )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-configuration-read-failed', self::EIO );
            }
            $currentiface = null;
            $currentifacefamily = null;
            // parse each line and generate internal representation
            foreach( $lines as $line )
            {
                $line = trim( $line );
                $_info = explode( ' ', $line );

                if( empty( $line ) || $line[0] == '#' || $_info[1] == 'lo' )
                {
                    continue;
                }
                else
                {
                    switch( $_info[0] )
                    {
                        case 'iface':
                            $currentiface = $_info[1];
                            if( !isset( $this->interfaces[$currentiface] ) )
                            {
                                $this->interfaces[$currentiface] = array( );
                            }
                            $currentifacefamily = $_info[2];
                            // @formatter:off
                            $this->interfaces[$currentiface][$currentifacefamily] = array(
                                'properties' => array( 'family' => $_info[2], 'method' => $_info[3] ),
                                'config' => array( ),
                                'immutable' => array( )
                            );
                            // @formatter:on
                            break;
                        // resolvconf(8) stanzas
                        case 'dns-nameservers':
                        case 'dns-search':
                            $field = $_info[0];
                            unset( $_info[0] );
                            $this->interfaces[$currentiface][$currentifacefamily]['config'][$field] = implode( ' ', $_info );
                            break;
                        case 'pre-up':
                        case 'up':
                        case 'post-up':
                        case 'pre-down':
                        case 'down':
                        case 'post-down':
                            $field = $_info[0];
                            unset( $_info[0] );
                            $this->interfaces[$currentiface][$currentifacefamily]['immutable'][$field] = implode( ' ', $_info );
                            break;
                        // stop filling the current interface
                        case 'mapping':
                        case 'auto':
                        case 'allow-hotplug':
                        case 'allow-auto':
                            if( !is_null( $currentiface ) )
                            {
                                $currentiface = null;
                            }
                            break;
                        default:
                            if( !is_null( $currentiface ) )
                            {
                                $_family = $this->interfaces[$currentiface][$currentifacefamily]['properties']['family'];
                                $_method = $this->interfaces[$currentiface][$currentifacefamily]['properties']['method'];
                                $field = $_info[0];

                                if( array_key_exists( $_family, self::$stanzas ) && array_key_exists( $_method, self::$stanzas[$_family] ) && array_key_exists( $field, self::$stanzas[$_family][$_method] ) )
                                {
                                    $field = $_info[0];
                                    unset( $_info[0] );
                                    $this->interfaces[$currentiface][$currentifacefamily]['config'][$field] = implode( ' ', $_info );
                                }
                                unset( $_family );
                                unset( $_method );
                            }
                            break;
                    }
                }
            }
        }

        /**
         * Save the configuration
         */
        public function save( )
        {
            if( count( $this->interfaces ) == 0 )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-invalid-config', self::EINVALID );
            }

            $_tmpname = self::INTERFACES_FILE . '.in';
            $_tmphandler = null;

            try
            {
                if( ( $_tmphandler = fopen( $_tmpname, 'w' ) ) === false )
                {
                    throw new KGatewayNetworkException( 'gateway-network-interface-error-write-failed', self::EIO );
                }

                if( ( $header = file_get_contents( self::INTERFACES_FILE . '.header' ) ) === false )
                {
                    throw new KGatewayNetworkException( 'gateway-network-interface-error-write-failed', self::EIO );
                }

                $date = date( 'Y-m-d H:i:s' );
                $this->writeLine( $_tmphandler, "# Last update: {$date}" );
                unset( $date );
                $this->writeLine( $_tmphandler, $header );

                // Check for dirty interface configurations and required stanzas
                foreach( $this->interfaces as $interface => &$data )
                {
                    $this->writeLine( $_tmphandler, '' );
                    $this->writeLine( $_tmphandler, "auto {$interface}" );
                    $this->writeLine( $_tmphandler, "allow-hotplug {$interface}" );

                    foreach( $data as $family => &$config )
                    {
                        if( strpos( $family, 'inet' ) === false )
                        {
                            continue;
                        }

                        $method = $config[ 'properties' ][ 'method' ];

                        $this->writeLine( $_tmphandler, '' );
                        $this->writeLine( $_tmphandler, "iface {$interface} {$family} {$method}" );

                        if( is_array( $config['config'] ) )
                        {
                            foreach( $config['config'] as $key => $value )
                            {
                                if( !empty( $value ) )
                                {
                                    $this->writeLine( $_tmphandler, "    {$key} {$value}" );
                                }
                            }
                        }

                        // issue #14435
                        // always define post-up callback!
                        if( !isset( $config['immutable'] ) || !isset( $config['immutable']['post-up'] ) )
                        {
                            $config['immutable']['post-up'] = '/usr/sbin/kgateway-ifup';
                        }
                        if( !isset( $config['immutable'] ) || !isset( $config['immutable']['post-down'] ) )
                        {
                            $config['immutable']['post-down'] = '/usr/sbin/kgateway-ifup';
                        }

                        foreach( $config['immutable'] as $key => $value )
                        {
                            $this->writeLine( $_tmphandler, "    {$key} {$value}" );
                        }
                    }
                }

                if( fclose( $_tmphandler ) === false || rename( $_tmpname, self::INTERFACES_FILE ) === false )
                {
                    throw new KGatewayNetworkException( 'gateway-network-interface-error-write-failed', self::EIO );
                }
            }
            catch( Exception $e )
            {
                if( $_tmphandler )
                {
                    fclose( $_tmphandler );
                }
                if( file_exists( $_tmpname ) )
                {
                    unlink( $_tmpname );
                }
                trigger_error( $e->getTraceAsString( ), E_USER_WARNING );
                throw $e;
            }
        }

        protected function write( $handler, $content )
        {
            if( fwrite( $handler, $content ) === false )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-write-failed', self::EIO );
            }
        }

        protected function writeLine( $handler, $content )
        {
            $this->write( $handler, "{$content}\n" );
        }

        /**
         * Get available interfaces
         *
         * @return array
         */
        public function getInterfaces( )
        {
            return array_keys( $this->interfaces );
        }

        /**
         * Remove a configured interface
         */
        public function removeInterface( $name )
        {
            if( $name === 'lo' )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-loopback', self::ENOPERM );
            }

            if( $this->hasInterface( $name ) )
            {
                unset( $this->interfaces[$name] );
            }
        }

        /**
         * Check if interface exists
         *
         * @param string $name Interface name
         */
        public function hasInterface( $name )
        {
            return isset( $this->interfaces[$name] );
        }

        /**
         * Get all defined DNS servers
         *
         * @return array
         */
        public function getDNSServers( )
        {
            $servers = array( );
            foreach( $this->interfaces as $intf => &$config )
            {
                if( $intf === 'lo' )
                {
                    continue;
                }

                if( isset( $config['inet']['config']['dns-nameservers'] ) )
                {
                    $servers = array_merge( $servers, explode( ' ', $config['inet']['config']['dns-nameservers'] ) );
                }
                if( isset( $config['inet6']['config']['dns-nameservers'] ) )
                {
                    $servers = array_merge( $servers, explode( ' ', $config['inet6']['config']['dns-nameservers'] ) );
                }
            }
            return array_unique( $servers );
        }

        /**
         * Get first defined DNS search domain
         *
         * @return string
         */
        public function getDNSSearchDomain( )
        {
            foreach( $this->interfaces as $intf => &$config )
            {
                if( $intf === 'lo' )
                {
                    continue;
                }

                if( isset( $config['inet']['config']['dns-search'] ) )
                {
                    return $config['inet']['config']['dns-search'];
                }
                if( isset( $config['inet6']['config']['dns-search'] ) )
                {
                    return $config['inet6']['config']['dns-search'];
                }
            }
        }

        /**
         * Get interface properties
         */
        public function getProperties( $interface, $family )
        {
            if( array_key_exists( $interface, $this->interfaces ) && array_key_exists( $family, $this->interfaces[ $interface ] ) )
            {
                return $this->interfaces[$interface]['properties'];
            }
            else
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-not-found', self::ERUNTIME );
            }
        }

        /**
         * Set interface properties
         */
        public function setProperties( $interface, $properties, $config )
        {
            if( $interface == 'lo' )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-loopback', self::ENOPERM );
            }
            elseif( !is_array( $properties ) )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-invalid-properties', self::EINVALID );
            }

            $this->setConfiguration( $interface, $config, $properties['method'], $properties['family'] );
        }

        /**
         * Get interface configuration
         */
        public function getConfiguration( $interface )
        {
            if( array_key_exists( $interface, $this->interfaces ) )
            {
                return $this->interfaces[$interface];
            }
            else
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-not-found', self::ERUNTIME );
            }
        }

        /**
         * Set interface configuration
         */
        public function setConfiguration( $interface, $config, $_method = 'dhcp', $_family = 'inet' )
        {
            if( $interface == 'lo' )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-loopback', self::ENOPERM );
            }
            elseif( !is_array( $config ) )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-invalid-config', self::EINVALID );
            }

            foreach( $config as $key => &$value )
            {
                // interfaces(5) stanzas
                if( array_key_exists( $key, self::$stanzas[$_family][$_method] ) )
                {
                    $this->check( $value, self::$stanzas[$_family][$_method][$key] );
                }
                // resolvconf(8) stanzas
                elseif( array_key_exists( $key, self::$stanzasResolvconf ) )
                {
                    $this->check( $value, self::$stanzasResolvconf[$key] );
                }
                // unknown stanza
                else
                {
                    throw new KGatewayNetworkException( 'gateway-network-interface-error-invalid-stanza', self::EINVALID );
                }
            }

            $this->interfaces[$interface][$_family]['properties']['method'] = $_method;
            $this->interfaces[$interface][$_family]['config'] = $config;

            if( !isset( $this->interfaces[$interface][$_family]['immutable'] ) || count( $this->interfaces[$interface][$_family]['immutable'] ) == 0 )
            {
                // add post-up and post-down if needed
                $this->interfaces[$interface][$_family]['immutable'] = array( 'post-up' => '/usr/sbin/kgateway-ifup' );
                $this->interfaces[$interface][$_family]['immutable'] = array( 'post-down' => '/usr/sbin/kgateway-ifup' );
            }
        }

        /**
         * Get the interface real mac address
         */
        public function getMacAddress( $interface )
        {
            if( array_key_exists( $interface, $this->interfaces ) )
            {
                $_f = "/sys/class/net/{$interface}/address";
                if( file_exists( $_f ) )
                {
                    return trim( file_get_contents( $_f ) );
                }
                trigger_error( "Unable to read MAC ADDRESS for interface {$interface}", E_USER_WARNING );
                return '';
            }
            else
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-not-found', self::ERUNTIME );
            }
        }

        // ********************** validation methods **************************
        /**
         * Check value using supplied configuration
         *
         * @param value string
         * @param configuration string|array|null
         *
         * @throws KGatewayNetworkException
         */
        protected function check( &$value, $configuration )
        {
            $validationMethod = null;
            $validationData = null;
            if( is_null( $configuration ) )
            {
                // no validator
                return;
            }
            elseif( is_array( $configuration ) )
            {
                $validationMethod = $configuration[0];
                $validationData = $configuration[1];
            }
            else
            {
                $validationMethod = $configuration;
            }

            if( !method_exists( $this, $validationMethod ) )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-validator-not-found', self::ERUNTIME );
            }
            else
            {
                if( is_null( $validationData ) )
                {
                    $ok = $this->$validationMethod( $value );
                }
                else
                {
                    $ok = $this->$validationMethod( $value, $validationData );
                }

                if( !$ok )
                {
                    trigger_error( "{$validationMethod}: failed", E_USER_NOTICE );
                    throw new KGatewayNetworkException( 'gateway-network-interface-error-invalid-value', self::EINVALID );
                }
            }
        }

        /**
         * Check if value is in ipv4 format
         *
         * data array valid options:
         *   empty => boolean - if true, value can be empty
         *   check-ebs-subnet => boolean - if true, check the reserved subnet
         *   restrict-first-and-last => boolean - if true, the first and the
         *       last octets should be between 1 and 254
         *
         * @param value string
         * @param data array options
         * @return boolean
         */
        public function checkIpv4( &$value, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $value ) )
            {
                return true;
            }
            // cpu <-> ebs network
            if( array_key_exists( 'check-ebs-subnet', $data ) && $data['check-ebs-subnet'] && substr( $value, 0, 10 ) == '172.31.249' )
            {
                throw new KGatewayNetworkException( 'gateway-network-interface-error-restricted-subnet', self::EINVALID );
            }
            $parts = explode( '.', $value );
            $rebuild = array( );
            $restrict = isset( $data['restrict-first-and-last'] ) && $data['restrict-first-and-last'];
            if( count( $parts ) == 4 )
            {
                for( $i = 0; $i < 4; $i++ )
                {
                    if( !is_numeric( $parts[$i] ) )
                    {
                        return false;
                    }

                    if( $restrict && ($i == 0 || $i == 3) )
                    {
                        if( (int)$parts[$i] < 1 || (int)$parts[$i] > 254 )
                        {
                            return false;
                        }
                    }
                    else
                    {
                        if( (int)$parts[$i] < 0 || (int)$parts[$i] > 255 )
                        {
                            return false;
                        }
                    }
                    $rebuild[] = (int)$parts[$i];
                }
                // rebuild address without leading zeros -- issue #4137
                $value = implode( '.', $rebuild );
                return true;
            }
            return false;
        }

        public function checkMacAddress( &$value, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $value ) )
            {
                return true;
            }
            return false;
        }

        public function checkFqdn( &$value, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $value ) )
            {
                return true;
            }
            // this regex was created using RFC 1035 as reference
            // http://tools.ietf.org/html/rfc1035#section-2.3.1
            // issue #13971
            return preg_match( '/([a-zA-Z][a-zA-Z0-9\-]+[a-zA-Z0-9]\.)+[a-zA-Z][a-zA-Z0-9]{1,2}/', $value ) === 1;
        }

        public function checkInteger( &$value, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $value ) )
            {
                return true;
            }
            if( is_numeric( $value ) )
            {
                if( array_key_exists( 'min', $data ) || array_key_exists( 'max', $data ) )
                {
                    $min = array_key_exists( 'min', $range ) ? $range['min'] : -PHP_INT_MAX;
                    $max = array_key_exists( 'max', $range ) ? $range['max'] : PHP_INT_MAX;
                    if( (int)$value < $min || (int)$value > $max )
                    {
                        return false;
                    }
                }
                return true;
            }
            return false;
        }

        public function checkAlwaysInvalid( &$value, $data = null )
        {
            return false;
        }

        public function checkIpv4OrIpv6( &$value, $data = array() )
        {
            return $this->checkIpv4( $value, $data ) || $this->checkIpv6( $value, $data );
        }

        public function checkIpxFrame( &$value, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $value ) )
            {
                return true;
            }
            return false;
        }

        public function checkIpxNetNum( &$value, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $value ) )
            {
                return true;
            }
            return false;
        }

        public function checkIpv6( &$value, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $value ) )
            {
                return true;
            }

            if( function_exists( 'filter_var' ) )
            {
                $validIPV6 = filter_var( $value, FILTER_VALIDATE_IP, FILTER_FLAG_IPV6 ) == false ? false : true;
            }
            //elastix compability
            else
            {
                $patternIPV6 = '/^((([0-9A-Fa-f]{1,4}:){7}[0-9A-Fa-f]{1,4})|(([0-9A-Fa-f]{1,4}:){6}:[0-9A-Fa-f]{1,4})|(([0-9A-Fa-f]{1,4}:){5}:([0-9A-Fa-f]{1,4}:)?[0-9A-Fa-f]{1,4})|(([0-9A-Fa-f]{1,4}:){4}:([0-9A-Fa-f]{1,4}:){0,2}[0-9A-Fa-f]{1,4})|(([0-9A-Fa-f]{1,4}:){3}:([0-9A-Fa-f]{1,4}:){0,3}[0-9A-Fa-f]{1,4})|(([0-9A-Fa-f]{1,4}:){2}:([0-9A-Fa-f]{1,4}:){0,4}[0-9A-Fa-f]{1,4})|(([0-9A-Fa-f]{1,4}:){6}((\b((25[0-5])|(1\d{2})|(2[0-4]\d)|(\d{1,2}))\b)\.){3}(\b((25[0-5])|(1\d{2})|(2[0-4]\d)|(\d{1,2}))\b))|(([0-9A-Fa-f]{1,4}:){0,5}:((\b((25[0-5])|(1\d{2})|(2[0-4]\d)|(\d{1,2}))\b)\.){3}(\b((25[0-5])|(1\d{2})|(2[0-4]\d)|(\d{1,2}))\b))|(::([0-9A-Fa-f]{1,4}:){0,5}((\b((25[0-5])|(1\d{2})|(2[0-4]\d)|(\d{1,2}))\b)\.){3}(\b((25[0-5])|(1\d{2})|(2[0-4]\d)|(\d{1,2}))\b))|([0-9A-Fa-f]{1,4}::([0-9A-Fa-f]{1,4}:){0,5}[0-9A-Fa-f]{1,4})|(::([0-9A-Fa-f]{1,4}:){0,6}[0-9A-Fa-f]{1,4})|(([0-9A-Fa-f]{1,4}:){1,7}:))$/';
                $validIPV6 = preg_match( $patternIPV6, $value ) ? true : false;
            }
            if( $validIPV6 )
            {
                return true;
            }

            return false;
        }

        public function checkIpv6Mask( &$value, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $value ) )
            {
                return true;
            }

            return ((int)$value > 128 || (int)$value <= 0) ? false : true;
        }

        /**
         * Check if all elements from the list are ipv4 or ipv6 strings, also
         * convert the value to native representation, if needed
         *
         * @param array|string list values list
         * @return boolean
         */
        public function checkIpv4OrIpv6List( &$list, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $list ) )
            {
                return true;
            }
            if( is_array( $list ) )
            {
                $l = &$list;
            }
            else
            {
                $l = explode( ' ', $list );
            }

            if( count( $l ) == 0 )
            {
                return false;
            }

            foreach( $l as &$v )
            {
                if( !$this->checkIpv4OrIpv6( $v, $data ) )
                {
                    return false;
                }
            }
            if( is_array( $list ) )
            {
                $_l = implode( ' ', $list );
                $list = $_l;
            }
            return true;
        }

    }
