(function (Khomp) {
    /**
    Decodificação do protocolo ISDN
    Webanalisador de log de mensagens Q.931

    Documentação auxiliar: http://devserver/svn/API/docs/Interno/ISDN/Normas/Normas%20ITU-T/

    @module isdn
    @submodule isdn-decoder
    @class ISDN
     */
    var ISDN = Khomp.ISDN = {

        logObj : null,
        logSource : '',
        messagesSuccess : {},
        busyChannels : {},
        usedChannels : {},
        /// Total de mensagens que foram apresentadas sem uma mensagem de SETUP. Isso indica que a mensagem de SETUP deve estar em outro arquivo de log mais antigo.
        totalMessagesWithoutSetup : 0

    };

    /**
    Limpa dos dados para uma novo processamento.

    @method clear
    @static
     */
    ISDN.clear = function () {

        this.logObj = null;
        this.logSource = '';
        this.messagesSuccess = {};
        this.totalMessagesWithoutSetup = 0;

        this.usedChannels = {};
        this.busyChannels = {};
    };

    /**
    Inicia o processamento parcial ou total do arquivo de log.
    Para iniciar uma nova análise deve ser chamado o método clear.

    Feito com base no formato de mensagens explicado na pg 47 da ITU Q.931

    @method extractData
    @static
    @param {String} logSource Conteúdo do arquivo isdn.log
    @return {Object} Log object com todos os dados das chamadas e mensagens

     */
    ISDN.extractData = function (logSource, callback) {

        console.time('TOTAL extractData:');

        callback = callback || function () {};

        logSource = (typeof logSource === 'string') ? logSource.split('\n') : logSource;

        var _this = this;
        var call;
        var callId;
        var line;
        var Message = this.Message;
        var messageObj;
        var InformationElements = this.InformationElements;

        console.log('Processando arquivo...', logSource.length);

        var logObj = this.logObj = (this.logObj !== null) ? this.logObj : {
            calls : [],
            messageTypes : {},
            links : {}
        };

        for (var i = 0, lines_len = logSource.length; i < lines_len; i++) {

            messageObj = new Message(logSource[i]);

            if (messageObj && messageObj.isValid !== false) {

                /// Para ter referência da linha do log
                messageObj.logLine = i;

                var deviceId = messageObj.getDeviceId();
                var datetime = messageObj.getDatetime();
                var type = messageObj.getType();
                var type_name = messageObj.getTypeName();
                var infos = messageObj.getInformationElements();
                var link = messageObj.getLink();
                var crv = messageObj.getCallReference();

                var key = deviceId + link + crv;

                call = this.busyChannels[key] || null;

                /// Considerar nova chamada quando a mensagem é SETUP (0x05)
                if (!call || type === 0x05) {

                    var crvDirection = messageObj.getCallReferenceDirection();

                    call = {
                        id : logObj.calls.length + 1,
                        Status : 'ERROR',
                        CallReference : crv,
                        Link : link,
                        Start : datetime,
                        // Nai : messageObj.nai,
                        deviceId : deviceId,
                        Direction : crvDirection,
                        // From : '',
                        // To : '',
                        Messages : []
                    };

                    logObj.calls.push(call);
                    this.busyChannels[key] = call;
                    logObj.links[link] = (logObj.links[link]) ? (logObj.links[link] + 1) : 1;

                }

                if (!call) {
                    // break;
                    continue;
                }
                
                callId = call.id;
                
                logObj.messageTypes[type_name] = (logObj.messageTypes[type_name]) ? logObj.messageTypes[type_name] + 1 : 1;

                _this.messagesSuccess[callId] = _this.messagesSuccess[callId] || {
                    'SETUP' : false,
                    'CONNECT' : false,
                    'DISCONNECT' : false,
                    'RELEASE' : false,
                    'RELEASE COMPLETE' : false
                };

                /// Vai associando a data da mensagem como sendo a última.
                call.Stop = datetime;

                call.Messages.push(messageObj);

                if (type_name in _this.messagesSuccess[callId]) {
                    _this.messagesSuccess[callId][type_name] = true;
                }
            }
        }

        callback();

        console.timeEnd('TOTAL extractData:');

        // this.logSource += logSource;

        return logObj;
    };

    /**
    Analisa cada uma das chamadas extraídas

    @method makeCallAnalysys
    @static
    @param {Function} callback
     */
    ISDN.makeCallAnalysys = function (callback) {

        /// Verifica se as chamadas têm erro a partir da análise das mensagens.
        console.time('TOTAL makeCallAnalysys:');

        var logObj = this.logObj;
        var calls = logObj.calls;
        var call;
        var messagesSuccess;

        logObj.totalCalls = 0;
        logObj.totalErrors = 0;
        logObj.totalAlerts = 0;
        logObj.totalSuccess = 0;
        logObj.errorCalls = [];
        logObj.alertCalls = [];
        logObj.firstCall = null;
        logObj.lastCall = null;
        logObj.messagesLengthByCall = [];

        callback = callback || function () {};

        for (var callPos = 0, calls_len = calls.length; callPos < calls_len; callPos++) {

            logObj.totalCalls++;
            call = calls[callPos];
            var callId = call.id;

            if (!logObj.firstCall) {
                logObj.firstCall = call;
            }

            logObj.lastCall = call;
            
            var msgs_length = logObj.lastCall.Messages.length;

            logObj.messagesLengthByCall.push(msgs_length);
            messagesSuccess = this.messagesSuccess[callId] || {};

            if (messagesSuccess.SETUP === true && messagesSuccess.RELEASE === true && messagesSuccess['RELEASE COMPLETE'] === true) {

                /// SUCCESS
                if (messagesSuccess.CONNECT === true && messagesSuccess.DISCONNECT === true) {

                    logObj.totalSuccess++;

                    call.Status = 'OK';

                } else {

                    /// ALERT

                    logObj.totalAlerts++;
                    logObj.alertCalls.push(callId);

                    call.Status = 'ALERT';
                }
            } else {

                logObj.errorCalls.push(callId);
                logObj.totalErrors++;
                call.Status = 'ERROR';
            }
        }

        console.timeEnd('TOTAL makeCallAnalysys:');
        callback();
    };

    /**
    Identifica se o formato do log é antigo (anterior à versão 3).

    @method isOldFormat
    @static
    @param {Mixed} logSource Conteúdo do arquivo de log em texto ou lista.
    @return {Number
     */
    ISDN.isOldFormat = function (logSource) {

        logSource = (typeof logSource === 'string') ? logSource.split('\n') : logSource;
        var line;

        for (var i = 0, lines_len = logSource.length; i < lines_len; i++) {

            line = logSource[i];

            /// Confere a barra '/' depois do dia que não tem na linha do formato antigo.
            if (line.indexOf('T|') === 0 && line.substr(4, 1) !== '/') {
                return true;
            }

        }

        return false;
    };

    /**
    Extrai o número de A (From) e B (To) da chamada.

    @method getCallFromTo
    @static
     */
    ISDN.getCallFromTo = function (call) {

        call.From = '';
        call.To = '';

        var msgs = call.Messages;
        var infos;
        var info;

        for (var msg_pos = 0, msgs_len = msgs.length; msg_pos < msgs_len; msg_pos++) {

            infos = msgs[msg_pos].infos;
            if (!infos) {
                continue;
            }

            for (var info_pos = 0, infos_len = infos.length; info_pos < infos_len; info_pos++) {
                info = infos[info_pos];
                if (info[0] === 0x6c) {
                    call.From += this.InformationElements.decodeInfo(0x6c, info.slice(1))['Number digits'];
                }
                if (info[0] === 0x70) {
                    call.To += this.InformationElements.decodeInfo(0x70, info.slice(1))['Number digits'];
                }
            }
        }
    };

    /**
    Verifica se a chamada tem todas os tipos de mensagens desejados

    @method callHasAllThisMessages
    @static
    @param {Array} messageTypes Lista com o código ou texto do tipo da mensagem
    @param {Object} call
    @return {Boolean}
     */
    ISDN.callHasAllThisMessages = function (messageTypes, call) {

        messageTypes = (typeof messageTypes === 'string' || typeof messageTypes === 'number') ? [messageTypes] : messageTypes;

        var found = 0;
        var msgs = call.Messages;
        var type;
        var regex_number = /^\d+$/;

        for (var i = 0, types_len = messageTypes.length; i < types_len; i++) {

            type = (regex_number.test(messageTypes[i])) ? parseInt(messageTypes[i], 10) : messageTypes[i];

            for (var j = 0, msgs_len = msgs.length; j < msgs_len; j++) {

                if (typeof type === 'number') {
                    if (msgs[j].type === type) {

                        found++;
                        break;
                    }
                } else {

                    if (msgs[j].getTypeName() === type) {

                        found++;
                        break;
                    }
                }
            }
        }

        return (types_len === found) ? true : false;

    };

    /**
    Procura por chamadas que tem uma das infos desejadas

    @method getCallsHasOneOfThisMessageInfos
    @static
    @param {Object} infos
    @param {Function} fnMatch Função para refinar a busca.
    @return {Array} Lista de chamadas que têm uma  das infos passadas
     */
    ISDN.getCallsHasOneOfThisMessageInfos = function (infos, fnMatch) {

        infos = (typeof infos === 'string') ? [infos] : infos;

        var found = 0;
        var calls = this.logObj.calls;
        var call;
        var msgs;
        var msg_pos;
        var msg_infos;
        var info;
        var info_pos;
        var calls_found = [];

        for (var call_pos = 0, calls_len = calls.length; call_pos < calls_len; call_pos++) {

            call = calls[call_pos];
            call.sequence = call_pos;
            msgs = call.Messages;
            found = false;

            for (msg_pos = 0, msgs_len = msgs.length; msg_pos < msgs_len; msg_pos++) {

                msg_infos = msgs[msg_pos].infos;

                if (!msg_infos) {
                    continue;
                }

                for (var msg_info_pos = 0, msg_infos_len = msg_infos.length; msg_info_pos < msg_infos_len; msg_info_pos++) {

                    info = msg_infos[msg_info_pos];

                    for (info_pos = 0, infos_len = infos.length; info_pos < infos_len; info_pos++) {

                        if (infos[info_pos] === info[0]) {

                            if (typeof fnMatch === 'function' && !fnMatch(info)) {
                                continue;
                            }

                            found = true;
                            break;
                        }
                    }
                    if (found) {
                        break;
                    }
                }
                if (found) {
                    calls_found.push(call);
                    break;
                }
            }

        }

        return calls_found;

    };

    /**
    Procura por chamadas que tem mensagens que têm parte do source desejado

    @method getCallsByMessageSource
    @static
    @param {String} source Parte do source para procurar.
    @return {Array} Lista de chamadas
     */
    ISDN.getCallsByMessageSource = function (source) {

        var calls = this.logObj.calls;
        var matchCalls = [];
        var call;
        var msg;

        for (var i = 0, calls_len = calls.length; i < calls_len; i++) {
            call = calls[i];

            for (var j = 0, msgs_len = call.Messages.length; j < msgs_len; j++) {

                msg = call.Messages[j];

                if (msg.source.indexOf(source) !== -1) {
                    matchCalls.push(call);
                }
            }
        }
        return matchCalls;
    };

    /**
    Procura por chamadas que tem mensagens com diferentes devices ou links. Se encontrar alguma está ocorrendo algum erro na separação das mensagens por chamada.

    @method getCallsWithDifferentsDevicesOrLinks
    @static
    @return {Array} Lista de chamadas com problema
     */
    ISDN.getCallsWithDifferentsDevicesOrLinks = function () {

        var calls = this.logObj.calls;
        var matchCalls = [];
        var call;
        var msg;
        var link;
        var deviceId;

        for (var call_pos = 0, calls_len = calls.length; call_pos < calls_len; call_pos++) {

            call = calls[call_pos];
            deviceId = call.deviceId;
            link = call.Link;

            for (var j = 0, msgs_len = call.Messages.length; j < msgs_len; j++) {

                msg = call.Messages[j];

                if (deviceId !== msg.getDeviceId() || link !== msg.getLink()) {
                    matchCalls.push(call);
                    break;
                }
            }
        }

        return matchCalls;
    };

    /**
    Procura por chamadas que tem mensagens com CRV diferente da chamada (considerar envio ou recebimento). Se encontrar alguma está ocorrendo algum erro na separação das mensagens por chamada.

    @method getCallsDifferentCRV
    @static
    @return {Array} Lista de chamadas com problema
     */
    ISDN.getCallsDifferentCRV = function () {

        var calls = this.logObj.calls;
        var matchCalls = [];
        var call;
        var msg;
        var crv;

        for (var call_pos = 0, calls_len = calls.length; call_pos < calls_len; call_pos++) {

            call = calls[call_pos];
            crv = call.CallReference;

            for (var j = 0, msgs_len = call.Messages.length; j < msgs_len; j++) {

                msg = call.Messages[j];

                if (crv === msg.getCallReference() || crv === msg.getCallReferenceOther()) {
                    continue;
                }
                matchCalls.push(call);
                break;
            }
        }

        return matchCalls;
    };

    /**
    Feito com base no formato de mensagens explicado na pg 47 da ITU Q.931

    @method getCallsWithError
    @static
    @param {String} Conteúdo do arquivo isdn.log
    @return {Object} Log object com todos os dados das chamadas e mensagens
     */
    ISDN.getCallsWithError = function () {

        var messages = [];
        var calls = this.logObj.calls;
        var errorIds = this.logObj.errorCalls;

        for (var i in errorIds) {
            messages.push(calls[errorIds[i]]);
        }

        return messages;
    };

    /**
    Para testes mandar reprocessar a decodificação

    @method reprocess
    @static
    @return {Boolean}
     */
    ISDN.reprocess = function () {

        if (this.logSource) {
            var logSource = this.logSource;
            this.clear();
            this.extractData(logSource, function () {
                console.log('Processado!');
            });
            return true;
        }
        return false;
    };

})(Khomp);
