(function (Khomp) {

    var ISDN = Khomp.ISDN;

    /**
    Decodificação dos 'Information Elements'

    @module isdn
    @submodule isdn-decoder
    @class InformationElements
     */
    var InformationElements = ISDN.InformationElements = {

        /**
        Tipos conhecidos de information elements

        Tradução dos códigos extraída da ITU-T
        Complementado com: http://www.zytrax.com/tech/protocols/isdn/ie.htm

        @attribute types
        @final
         */
        types : {

            /// 1:::---- Single octet information elements (type a (1) e b (2))
            0x80 : 'Reserved', // 1[000]---- - type 1
            0x90 : 'Shift', // 1[001]---- - type 1

            0xa0 : 'More data', // 1[010]0000 - type 2
            0xa1 : 'Sending complete', // 1[010]0001 - type 2

            0xb0 : 'Congestion level', // 1[011]---- - type 1
            0xd0 : 'Repeat indicator', // 1[101]---- - type 1

            /// 0::::::: Variable length information element  (type c (3))
            0x00 : 'Segmented message',
            0x04 : 'Bearer capability',
            0x08 : 'Cause',
            0x0d : 'Extended facility',
            0x10 : 'Call identity',
            0x14 : 'Call state',
            0x18 : 'Channel identification',
            0x1c : 'Facility',
            0x1e : 'Progress indicator',
            0x20 : 'Network-specific facilities',
            0x27 : 'Notification indicator',
            0x28 : 'Display',
            0x29 : 'Date/time',
            0x2c : 'Keypad facility',
            0x32 : 'Information request',
            0x34 : 'Signal',
            0x38 : 'Feature activation',
            0x39 : 'Feature indication',
            0x3a : 'SPID - Service Profile Identification',
            0x3b : 'Endpoint identifier',
            0x40 : 'Information rate',
            0x42 : 'End-to-end transit delay',
            0x43 : 'Transit delay selection and indication',
            0x44 : 'Packet layer binary parameters',
            0x45 : 'Packet layer window size',
            0x46 : 'Packet size',
            0x47 : 'Closed user group',
            0x4a : 'Reverse charging indication',
            0x4c : 'Connected number',
            0x6c : 'Calling party number',
            0x6d : 'Calling party subaddress',
            0x70 : 'Called party number',
            0x71 : 'Called party subaddress',
            0x74 : 'Redirecting number',
            0x76 : 'Redirection number',
            0x78 : 'Transit network selection',
            0x79 : 'Restart indicator',
            0x7c : 'Low layer compatibility',
            0x7d : 'High layer compatibility',
            0x7e : 'User-user',
            0x7f : 'Escape for extension'
        },

        /**
        Se o bit é 0 (ZERO) continua, se é 1 (UM) é o último octeto
        Esta regra é utilizada e descrita detalhadamente na ITU Q.931

        @method isLastOctet
        @param {Number} num Octeto para verificar o bit de continuidade de informações
        @return {Boolean}
         */
        isLastOctet : function (num) {

            num = (typeof num === 'number') ? num : parseInt(num, 16);
            return ((num >> (7)) % 2 !== 0) ? true : false;
        },

        /**
        O elemento de informação pode ser representado de três formas (pg 52):
        Os três tipos são identificados como 1,2,3

        @method getOctetFormat
        @param elementCode
        @return {Number} Número associado ao tipo do elemento
         */
        getOctetFormat : function (elementCode) {

            var code = (typeof elementCode === 'string') ? parseInt(elementCode, 16) : elementCode;

            /// Se o primeiro bit é ZERO
            if ((code & 0x80) >> 7 === 0) {
                return 3;
            }

            /// Se o primeiro bit é UM
            if (code === 0xa1 || code === 0xa0) {
                return 2;
            }

            return 1;
        },

        /**
        Decodifica um Information element

        @method decodeInfo
        @param {Number} elementCode Código do Information element
        @param {Array} data Dados do Information element
        @return {Object} Decoded data
         */
        extractInfos : function (data) {

            var octetFormat;
            var elementCode;
            var infoSlice;
            var obj = {};
            var infos = [];
            var info;
            while (data.length) {

                elementCode = data.shift();

                octetFormat = this.getOctetFormat(elementCode);

                switch (octetFormat) {

                case 1:
                case 2:
                    infos.push([elementCode]);
                    // obj[elementCode] = null;
                    break;

                default:
                    infoSlice = data.shift();

                    info = data.splice(0, infoSlice);
                    info.unshift(elementCode);

                    infos.push(info);

                    // obj[elementCode] = data.splice(0, infoSlice);
                }
            }
            return infos;
            // return obj;
        },

        /**
        Decodifica um Information element

        @method decodeInfo
        @param {Number} elementCode Código do Information element
        @param {Array} data Dados do Information element
        @return {Object} Decoded data
         */
        decodeInfo : function (elementCode, data) {

            var key;
            var value;
            var elementName;

            var code = (typeof elementCode === 'number') ? elementCode : parseInt(elementCode, 16);

            /// Extension element (0x7f), see details in ITU Q.931
            if (code === 0x7f) {

                key = data[2];
                var length = data[1];

                return data.slice(3, length + 2);
            }

            var octetFormat = this.getOctetFormat(code);

            switch (octetFormat) {

            case 1:

                key = code;
                value = (code & 0x0f);

                /// TODO: Tratamento especial para quando Shift. Lock e Non-lock

                value = (key in this.elementsDecoders) ? this.elementsDecoders[key](value) : value;

                break;

            case 2:

                key = code;
                value = true;

                value = (key in this.elementsDecoders) ? this.elementsDecoders[key](value) : value;

                break;

            case 3:

                key = code;
                value = data.slice(0);

                if (!(key in this.elementsDecoders)) {
                    console.log('UNKNOW ELEMENT:', key, value);
                }

                value = (key in this.elementsDecoders) ? this.elementsDecoders[key](value) : value;

                break;

            default:
                return;
            }

            return value;
        },

        /**
        Atributos suportados pelos elementos
        Chave valor numérica para performance

        @attribute attributeNames
        @final
         */
        attributeNames : {
            0x00 : 'Additional layer 3 protocol information',
            0x01 : 'Assignor/assignee',
            0x02 : 'Call state value',
            0x03 : 'Causes',
            0x04 : 'Channel type/map element type',
            0x05 : 'Class',
            0x06 : 'Coding standard',
            0x07 : 'D-channel indicator',
            0x08 : 'Default packet size',
            0x09 : 'Extended audiovisual characteristics identification',
            0x0a : 'Extended high layer characteristics identification',
            0x0b : 'Flow control on Rx',
            0x0c : 'Flow control on Tx',
            0x0d : 'High layer characteristics identification',
            0x0e : 'In-band/out-band negotiation',
            0x0f : 'Information channel selection',
            0x10 : 'Information transfer capability',
            0x11 : 'Information transfer rate',
            0x12 : 'Interface identifier present',
            0x13 : 'Interface type',
            0x14 : 'Intermediate rate',
            0x15 : 'Interpretation',
            0x16 : 'Location',
            0x17 : 'Logical link identifier negotiation',
            0x18 : 'Mode duplex',
            0x19 : 'Mode of operation',
            0x1a : 'Mode of operation layer 2',
            0x1b : 'Mode of operation layer 3',
            0x1c : 'Modem type',
            0x1d : 'Multiframe',
            0x1e : 'Negotiation',
            0x1f : 'Negotiation indicator',
            0x20 : 'Network identification plan',
            0x21 : 'NIC on Rx',
            0x22 : 'NIC on Tx',
            0x23 : 'Notification description',
            0x24 : 'Number/map',
            0x25 : 'Number of data bits excluding parity Bit if present',
            0x26 : 'Number of stop bits',
            0x27 : 'Numbering plan',
            0x28 : 'Odd/even indicator',
            0x29 : 'Parity information',
            0x2a : 'Preferred/Exclusive',
            0x2b : 'Presentation indicator',
            0x2c : 'Presentation method of protocol profile',
            0x2d : 'Progress description',
            0x2e : 'Protocol discriminator',
            0x2f : 'Q.933 use',
            0x30 : 'Rate adaption header',
            0x31 : 'Screening indicator',
            0x32 : 'Signal value',
            0x33 : 'Synchronous/Asynchronous',
            0x34 : 'Transfer mode',
            0x35 : 'Type of network identification',
            0x36 : 'Type of number',
            0x37 : 'Type of subaddress',
            0x38 : 'User information layer 1 protocol',
            0x39 : 'User information layer 2 protocol',
            0x3a : 'User information layer 3 protocol',
            0x3b : 'User rate'
        },

        /**
        Atributos suportados pelos elementos
        Alguns podem ser reaproveitados entre os elementos

        @attribute attributes
         */
        attributes : {},

        /**
        Elementos suportados. Nesta é alimentada a lógica específica para cada tipo.
        Os itens estão separados em arquivos na pasta information_elements

        @attribute elementsDecoders
         */
        elementsDecoders : {}
    };

})(Khomp);
