<?php
    class KJsonRpcException extends Exception
    {
        /* VOID */
    }

    class KJsonRpcClient
    {
        protected $host;
        protected $port;
        protected $timeout;
        protected $connection;
        protected $requestId = 1;
        protected $configFile = null;

        /**
         * Create new rpc client. The connection will be lazy loaded until the
         * first
         * request.
         *
         * @param string $host Hostname, IP address or configuration name
         * @param int $port Host port, if null will use $host as configuration
         * name
         * @param int $timeout Timeout in seconds
         */
        public function __construct( $host, $port = null, $timeout = 10 )
        {
            unset( $this->connection );
            unset( $this->configFile );
            $this->host = $host;
            $this->port = $port;
            $this->timeout = $timeout;
        }

        /**
         * Lazy load some attributes
         *
         * @param string $name Parameter name
         * @return mixed
         */
        public function __get( $name )
        {
            switch( $name )
            {
                case 'connection':
                    if( is_null( $this->port ) )
                    {
                        $this->__readConfig( );
                    }
                    $this->connection = fsockopen( $this->host, $this->port );
                    if( $this->connection === FALSE )
                    {
                        throw new KJsonRpcException( 'Connection failed' );
                    }
                    if( $this->timeout )
                    {
                        stream_set_timeout( $this->connection, $this->timeout );
                    }
                    return $this->connection;
                    break;
                case 'configFile':
                    $app = KSystem::getInstance( );
                    $this->configFile = $app->getConfigDir( ) . DIRECTORY_SEPARATOR . 'jsonrpc.json';
                    return $this->configFile;
                    break;
            }
            throw new Exception( "Invalid attribute {$name}" );
        }

        /**
         * Call remote procedure
         *
         * @param string $method Remote procedure
         * @param array $params Parameters array
         */
        public function __call( $method, $params )
        {
            $request = new stdClass;
            $request->id = $this->requestId++;
            $request->jsonrpc = '2.0';
            $request->method = $method;
            $request->params = $params;

            $this->__send( $request );
            return $this->__recv( );
        }

        /**
         * Read configuration from file
         */
        protected function __readConfig( )
        {
            // json_decode does not support comments, so we need strip all of
            // them
            // REGEX from http://am1.php.net/json_decode#112735
            $data = json_decode( preg_replace( "#(/\*([^*]|[\r\n]|(\*+([^*/]|[\r\n])))*\*+/)|([\s\t]//.*)|(^//.*)#", '', file_get_contents( $this->configFile ) ) );
            if( is_null( $data ) )
            {
                throw new KJsonRpcException( 'Unable to read configuration file' );
            }
            foreach( $data->Handlers as &$handler )
            {
                if( $handler->Name == $this->host )
                {
                    trigger_error( "Using configuration handler {$this->host} to initialize JSON-RPC client class -- will connect to {$handler->Host}:{$handler->Port}", E_USER_NOTICE );
                    $this->host = $handler->Host;
                    $this->port = $handler->Port;
                    return;
                }
            }
            throw new KJsonRpcException( "Handler for {$this->host} not found in configuration" );
        }

        /**
         * Send request
         *
         * @param stdClass $request Request object
         */
        protected function __send( $request )
        {
            if( fwrite( $this->connection, json_encode( $request ) ) === FALSE || fflush( $this->connection ) === FALSE )
            {
                throw new KJsonRpcException( 'Send failed' );
            }
        }

        /**
         * Receive response
         */
        protected function __recv( )
        {
            $response = fgets( $this->connection );
            if( $response === FALSE )
            {
                throw new KJsonRpcException( 'Receive failed' );
            }
            $response = json_decode( $response );

            if( isset( $response->error ) )
            {
                throw new KJsonRpcException( $response->error->message, $response->error->code );
            }

            return $response->result;
        }

    }
