<?php
    class KLicenseManagerException extends Exception
    {
    }

    class KLicenseException extends Exception
    {
    }

    /**
     * Licenses managment
     *
     * TODO: convert this class to singleton
     */
    class KLicenseManager
    {
        /**
         * string Licenses directory
         */
        protected $licensesDir = null;
        /**
         * string Current date
         */
        protected $now = null;
        /**
         * string +15 days -- used to check if license will expire soon
         */
        protected $willexpire = null;
        /**
         * array[KLicense] licenses cache
         */
        protected $licenses = array( );
        /**
         * int Input/Output errors
         */
        const EIO = 1;
        /**
         * int Invalid data
         */
        const EINVALID = 2;

        public function __construct( )
        {
            $app = KSystem::getInstance( );
            if( PHP_OS == 'WINNT' )
            {
                $this->licensesDir = "{$app->getInstallDir()}\\licenses";
            }
            else
            {
                $this->licensesDir = "{$app->getInstallDir()}/licenses";
            }
            if( !is_dir( $this->licensesDir ) )
            {
                $this->licensesDir = null;
                throw new KLicenseManagerException( 'license-manager-error-directory-not-found', self::EIO );
            }
            $this->now = date( 'Y-m-d' );
            // +15 days
            $this->willexpire = date( 'Y-m-d', time( ) + (15 * 86400) );
        }

        /**
         * Check if any license has the required type
         *
         * @see KLicense::TYPE_* for valid license types
         *
         * @param int type Required type
         * @return boolean
         */
        public function anyLicenseHasType( $type )
        {
            foreach( $this->getLicenses() as $license )
            {
                if( $license->hasType( $type ) && $license->isValid( ) )
                {
                    return true;
                }
            }
            return false;
        }

        /**
         * Check if any license has any of the required type
         *
         * @see KLicense::TYPE_* for valid license types
         *
         * @param int type Required type
         * @return boolean
         */
        public function anyLicenseHasAnyType( $type )
        {
            foreach( $this->getLicenses() as $license )
            {
                if( $license->hasAnyType( $type ) && $license->isValid( ) )
                {
                    return true;
                }
            }
            return false;
        }

        /**
         * Sum $type field from all valid licenses
         *
         * @see KLicense::TYPE_* for valid field types
         *
         * @param int type Required type
         * @param int serial Licensed device serial number
         * @return int Sum of all valid licenses that has the field
         */
        public function sumLicensesFieldValue( $type, $serial = null )
        {
            $total = 0;
            foreach( $this->getLicenses() as $license )
            {
                if( !is_null( $serial ) && $license->serial_number != $serial )
                {
                    continue;
                }
                if( $license->hasType( $type ) && $license->isValid( ) )
                {
                    $total += (int)$license->get( $license->getTypeAttributeName( $type ) );
                }
            }
            return $total;
        }

        /**
         * Calculate new value of $type field from all valid licenses + a new license
         *
         * @see KLicense::TYPE_* for valid field types
         *
         * @param int type Required type
         * @param KLicense new license
         * @return int calculated value of all valid licenses that has the field
         */
        public function calculateLicensesFieldValue( $type, $newlicense, $serial = null )
        {
            if ( ( $newlicense instanceof KLicense ) === false )
            {
                return 0;
            }

            $total = 0;

            if ( $newlicense->hasType( $type ) )
            {
                $total += (int)$newlicense->get( $newlicense->getTypeAttributeName( $type ) );
            }

            foreach( $this->getLicenses() as $license )
            {
                if( $newlicense->id == $license->id )
                {
                    continue;
                }
                if( $license->hasType( $type ) && $license->isValid( ) )
                {
                    if ( is_null( $serial ) || $license->serial_number === $serial )
                    {
                        $total += (int)$license->get( $license->getTypeAttributeName( $type ) );
                    }
                }
            }

            return $total;
        }

        /**
         * Scan license directory and return an array of licenses objects
         *
         * @return array of licenses
         */
        public function getLicenses( )
        {
            if( count( $this->licenses ) == 0 )
            {
                if( PHP_OS == 'WINNT' )
                {
                    $files = glob( "{$this->licensesDir}\\*.klf" );
                }
                else
                {
                    $files = glob( "{$this->licensesDir}/*.klf" );
                }
                if( $files === false || !is_array( $files ) )
                {
                    throw new KLicenseManagerException( 'license-manager-error-directory-read-failed', self::EIO );
                }
                elseif( count( $files ) > 0 )
                {
                    foreach( $files as $file )
                    {
                        try
                        {
                            $this->licenses[] = $this->getLicense( basename( $file ) );
                        }
                        catch( KLicenseException $e )
                        {
                            trigger_error( $e->getMessage( ), E_USER_WARNING );
                        }
                    }
                }
            }
            return $this->licenses;
        }

        /**
         * Get license object
         *
         * @param name string License (file) name
         * @return KLicense
         */
        public function getLicense( $name )
        {
            // if has cache, search it first
            if( count( $this->licenses ) > 0 )
            {
                foreach( $this->licenses as $license )
                {
                    if( $license->getFilename( ) == $name )
                    {
                        return $license;
                    }
                }
            }
            return new KLicense( $this, $name );
        }

        /**
         * Get license as string
         *
         * @param name string License (file) name
         * @return string
         */
        public function getLicenseRaw( $name )
        {
            return file_get_contents( KLicense::getFilenameWithFullPath( $this, $name ) );
        }

        /**
         * Delete license from licenses directory
         *
         * @param name string License (file) name
         * @param force boolean Force license file removal
         */
        public function deleteLicense( $name, $force = false )
        {
            // if is gateway (KMG or KMG SERVER), then check if the device is
            // being used by any NAP
            if( IS_GATEWAY && !$force )
            {
                require_once KCONFIG . '/classes/KGatewayYaml.php';

                $license = $this->getLicense( $name );
                $kgateway = new KGatewayYaml( );
                if( $kgateway->checkIfDeviceIsBeingUsedByAnyNap( $license->serial_number ) )
                {
                    throw new KLicenseException( 'license-manager-error-device-in-use-by-nap', self::EINVALID );
                }
                else
                {
                    $removeDevice = true;
                    foreach( $this->getLicenses() as $_license )
                    {
                        if( $license->id === $_license->id )
                        {
                            continue;
                        }
                        elseif( $license->serial_number === $_license->serial_number )
                        {
                            $removeDevice = false;
                            break;
                        }
                    }
                    if( $removeDevice )
                    {
                        global $sistema;
                        $sistema->deleteDevice( $license->serial_number );
                    }
                }
            }
            if( unlink( KLicense::getFilenameWithFullPath( $this, $name ) ) === false )
            {
                throw new KLicenseManagerException( 'license-manager-error-unlink-failed', self::EIO );
            }
            // reset cache
            $this->clearCache( );
        }

        /**
         * Clear lincense cache
         */
        public function clearCache( )
        {
            $this->licenses = array( );
        }

        /**
         * Upload license to licenses directory
         *
         * @param info array Information array (as part of $_FILE super-global)
         */
        public function uploadLicense( $info )
        {
            // .klt
            $_tmpname = substr( $info['name'], 0, strlen( $info['name'] ) - 1 ) . 't';
            $_tmpfilename = KLicense::getFilenameWithFullPath( $this, $_tmpname, false );
            if( move_uploaded_file( $info['tmp_name'], $_tmpfilename ) === false )
            {
                throw new KLicenseManagerException(
                    'license-manager-error-upload-failed',
                    self::EIO );
            }

            try
            {
                $newlicense = new KLicense( $this, $_tmpname );
                if( !$newlicense->hasSignature( ) )
                {
                    throw new KLicenseException(
                        'license-manager-error-signature-not-found',
                        self::EINVALID );
                }

                $_name = "{$newlicense->id} - K{$newlicense->serial_number} - {$newlicense->getTypeAsString( ' ', '%v %n' )} - {$newlicense->customer_info}.klf";
                $_filename = KLicense::getFilenameWithFullPath( $this, $_name, false );
                $currentlicense = null;
                foreach( $this->getLicenses() as $license )
                {
                    if( $newlicense->id == $license->id )
                    {
                        $currentlicense = $license;
                        break;
                    }
                }

                // if license already exists, then check if the new has a newer
                // issue_date
                if( !is_null( $currentlicense ) )
                {
                    if( $newlicense->getIssued( ) < $currentlicense->getIssued( ) )
                    {
                        throw new KLicenseManagerException(
                            'license-manager-error-current-license-is-newer',
                            self::EINVALID );
                    }
                }

                if( IS_GATEWAY )
                {
                    require_once KCONFIG . '/classes/KGatewayYaml.php';
                    require_once KROOT . '/classes/KLicenseLimits.php';

                    $kgateway = new KGatewayYaml( );

                    $saveyaml = false;

                    try
                    {
                        $limits = new KLicenseLimits( );

                        $hasE1 = false;
                        if ( $newlicense->hasType( KLicense::TYPE_GATEWAYLINKS ) )
                        {
                            $serial_links = (int)$this->calculateLicensesFieldValue(
                                KLicense::TYPE_GATEWAYLINKS,
                                $newlicense,
                                $newlicense->serial_number );

                            $limits->checkEBSLimits( "e1", $serial_links );

                            $hasE1 = true;
                        }
                        $hasGSM = false;
                        if ( $newlicense->hasType( KLicense::TYPE_GATEWAYGSMCHANNELS ) )
                        {
                            $serial_channels = (int)$this->calculateLicensesFieldValue(
                                KLicense::TYPE_GATEWAYGSMCHANNELS,
                                $newlicense,
                                $newlicense->serial_number );

                            $limits->checkEBSLimits( "gsm", $serial_channels );

                            $hasGSM = true;
                        }

                        $voip_calls =
                            (int)$this->calculateLicensesFieldValue(
                                KLicense::TYPE_GATEWAYVOIPCALLS,
                                $newlicense );
                        $e1_links =
                            (int)$this->calculateLicensesFieldValue(
                                KLicense::TYPE_GATEWAYLINKS,
                                $newlicense );
                        $gsm_channels =
                            (int)$this->calculateLicensesFieldValue(
                                KLicense::TYPE_GATEWAYGSMCHANNELS,
                                $newlicense );

                        $limits->checkGatewayLimits(
                            $e1_links * 30,
                            $gsm_channels,
                            $voip_calls,
                            $kgateway );

                        $limits->checkEBSE1LinkLimits( $e1_links );

                        if ( $hasGSM || $hasE1 )
                        {
                            require_once KCONFIG . '/classes/KDevicesYaml.php';

                            $kyaml = new KDevicesYaml( );
                            $yaml = $kyaml->getYaml( );

                            if( isset( $yaml['Devices'] ) === false )
                            {
                                $yaml['Devices'] = array();
                            }

                            $dev = null;

                            foreach( $yaml['Devices'] as &$device )
                            {
                                if( isset( $device[$newlicense->serial_number] ) )
                                {
                                    $dev = &$device[$newlicense->serial_number];

                                    if ( $hasE1 )
                                    {
                                        $serial_links = (int)$this->calculateLicensesFieldValue(
                                            KLicense::TYPE_GATEWAYLINKS,
                                            $newlicense,
                                            $newlicense->serial_number );

                                        $currentlinks = count( $dev['Links'] );

                                        if( $currentlinks < $e1_links )
                                        {
                                            for( $l = $currentlinks; $l < $serial_links; $l++ )
                                            {
                                                $dev['Links'][] = array(
                                                    (string)$l => array(
                                                        'Type' => 'E1',
                                                        'Name' => '',
                                                        'Profile' => 'ISDN'
                                                    )
                                                );
                                            }

                                            $saveyaml = true;
                                        }
                                    }

                                    break;
                                }
                            }

                            if ( is_null( $dev ) )
                            {
                                global $sistema;

                                $app = KSystem::getInstance( );

                                // Check total supported devices -- issue #3504
                                if( $limits->getEBSLimit() <= count( $yaml['Devices'] ) )
                                {
                                    throw new KLicenseException(
                                        'license-manager-error-maximum-supported-devices',
                                        self::EINVALID );
                                }

                                $type = null;
                                $version = null;

                                $kibsdevices = $sistema->getNotInstalledDevices( 0 );

                                foreach( $kibsdevices as &$device )
                                {
                                    if( $device[0] == $newlicense->serial_number )
                                    {
                                        $type = $device[1];
                                        $version = $device[5];

                                        break;
                                    }
                                }

                                if( is_null( $type ) || is_null( $version ) )
                                {
                                    throw new KLicenseException(
                                        'license-manager-error-connect-first',
                                        self::EINVALID );
                                }

                                $dev = array(
                                    'Type' => $type,
                                    'Version' => $version,
                                    'IP' => '',
                                    'Links' => array(),
                                    'Replacer' => '',
                                    'CustomTonesProfile' => 'Profile1'
                                );

                                if( $hasE1 )
                                {
                                    $serial_links = (int)$this->calculateLicensesFieldValue(
                                        KLicense::TYPE_GATEWAYLINKS,
                                        $newlicense,
                                        $newlicense->serial_number );

                                    for( $l = 0; $l < $serial_links; $l++ )
                                    {
                                        $dev['Links'][] = array(
                                            (string)$l => array(
                                                'Type' => 'E1',
                                                'Name' => '',
                                                'Profile' => 'ISDN'
                                            )
                                        );
                                    }
                                }
                                else
                                {
                                    //All GSM devices are configured as 16 channels
                                    $dev['ChannelGroups'][] = array(
                                        '0-15' => array(
                                            'Type' => 'GSM',
                                            'Profile' => 'GSM'
                                        )
                                    );
                                }

                                $yaml['Devices'][][$newlicense->serial_number] = $dev;

                                $saveyaml = true;
                            }
                        }
                    }
                    catch( KLicenseLimitsException $e )
                    {
                        if ( unlink( $_tmpfilename ) === false )
                        {
                            trigger_error( "Could not unlink file: " . $_tmpfilename, E_USER_WARNING );
                        }
                        throw new KLicenseManagerException( $e->getMessage(), self::EINVALID );
                    }

                    if( $saveyaml )
                    {
                        $kyaml->setYaml( $yaml );
                        $kyaml->selfAssignEbsIpAddresses( );
                        $kyaml->saveYaml( true );
                    }
                }

                // commit new license
                if( rename( $_tmpfilename, $_filename ) === false )
                {
                    throw new KLicenseManagerException(
                        'license-manager-error-rename-failed',
                        self::EIO );
                }

                $_tmpname = null;

                // if their names are different, remove the "current"
                if( !is_null( $currentlicense ) && $currentlicense->getFilename( ) != $_name )
                {
                    // deleteLicense already clear the licenses cache
                    $this->deleteLicense( $currentlicense->getFilename( ), true );
                }
                else
                {
                    // clear the licenses cache to force re-read all licenses
                    $this->clearCache( );
                }
            }
            // gambi finally, rah!
            catch( Exception $e )
            {
                try
                {
                    if( !is_null( $_tmpname ) )
                    {
                        $this->deleteLicense( $_tmpname, true );
                    }
                }
                catch( Exception $_e )
                {
                    /* ignore errors */
                }
                throw $e;
            }

            // reset cache
            $this->clearCache( );
        }

        /**
         * Get current date
         *
         * @return string
         */
        public function getNow( )
        {
            return $this->now;
        }

        /**
         * Get date +15 days
         *
         * @return string
         */
        public function getWillExpire( )
        {
            return $this->willexpire;
        }

        /**
         * Get licenses directory
         */
        public function getLicensesDirectory( )
        {
            return $this->licensesDir;
        }

    }

    /**
     * License object
     */
    class KLicense
    {
        // string License file name
        protected $filename = null;
        // KLicenseManager manager
        protected $manager = null;
        // array License data
        protected $license = array( );
        // int License type
        protected $type = 0;
        // array Expires date as array
        protected $expires = null;
        // array Issued date as array
        protected $issued = null;
        // true if this license has signature
        protected $signatureFound = false;

        // int Input/Output error
        const EIO = 1;
        // int Invalid license error
        const EINVALID = 2;

        // Types from KLicense.cpp
        // int Unknown type
        const TYPE_NONE = 0;
        // int FAX_CHANNELS
        const TYPE_FAX = 1;
        // int ISUP_CIRCUITS
        const TYPE_ISUP = 2;
        // int MTP2_LINKS
        const TYPE_MTP2 = 4;
        // int MTP_DUAL_MODE
        const TYPE_MTP = 8;
        // int G729_CHANNELS (deprecated since issue #5648)
        const TYPE_G729 = 0x10;
        // int VOIP_CHANNELS
        const TYPE_VOIP = 0x20;
        // int GATEWAY_ENABLE
        const TYPE_GATEWAYENABLE = 0x40;
        // int GATEWAY_LINKS
        const TYPE_GATEWAYLINKS = 0x80;
        // int GATEWAY_GSM_CHANNELS
        const TYPE_GATEWAYGSMCHANNELS = 0x100;
        // int PLAY_CHANNELS
        const TYPE_PLAYCHANNELS = 0x200;
        // int RECORD_CHANNELS
        const TYPE_RECORDCHANNELS = 0x400;
        // int GATEWAY_ANALYTICS
        const TYPE_GATEWAYANALYTICS = 0x800;
        // int GATEWAY_ANALYTICS_TRUNK0
        const TYPE_GATEWAYANALYTICSTRUNK0 = 0x1000;
        // int GATEWAY_ANALYTICS_TRUNK1
        const TYPE_GATEWAYANALYTICSTRUNK1 = 0x2000;
        // int GATEWAY_ANALYTICS_TRUNK2
        const TYPE_GATEWAYANALYTICSTRUNK2 = 0x4000;
        // int GATEWAY_ANALYTICS_TRUNK3
        const TYPE_GATEWAYANALYTICSTRUNK3 = 0x8000;
        // int GATEWAY_ANALYTICS_VOIP_CALLS
        const TYPE_GATEWAYANALYTICSVOIPCALLS = 0x10000;
        // int GATEWAY_VOIP_CALLS
        const TYPE_GATEWAYVOIPCALLS = 0x20000;

        // const analytics types
        const TYPES_GATEWAY_ANALYTICS = 0x1F800;

        public function __construct( KLicenseManager $manager, $name )
        {
            $this->manager = $manager;
            $this->filename = $name;

            // parse license
            $data = file( self::getFilenameWithFullPath( $manager, $name ), FILE_IGNORE_NEW_LINES );
            if( $data[0] != 'KHOMP LICENSE' )
            {
                $this->filename = null;
                throw new KLicenseException( 'license-error-invalid', self::EINVALID );
            }
            unset( $data[0] );
            foreach( $data as $line )
            {
                // end of license data
                if( strlen( $line ) > 0 && $line[0] === chr( 27 ) )
                {
                    $this->signatureFound = true;
                    break;
                }
                $line = trim( $line );
                // ignore empty lines
                if( strlen( $line ) > 0 )
                {
                    $keypos = strpos( $line, '=' );
                    $key = strtolower( substr( $line, 0, $keypos ) );
                    // issue 5648
                    if( $key === 'g729_channels' )
                    {
                        trigger_error( "License {$this->filename} has deprecated key {$key}. Key ignored.", E_USER_NOTICE );
                        continue;
                    }
                    $value = substr( $line, $keypos + 1 );
                    $value = is_string( $value ) ? $value : '';
                    $this->license[$key] = $value;
                }
            }

            if( $this->hasKey( 'fax_channels' ) )
            {
                $this->type |= self::TYPE_FAX;
            }
            if( $this->hasKey( 'isup_circuits' ) )
            {
                $this->type |= self::TYPE_ISUP;
            }
            if( $this->hasKey( 'mtp2_links' ) )
            {
                $this->type |= self::TYPE_MTP2;
            }
            if( $this->hasKey( 'mtp_dual_mode' ) )
            {
                $this->type |= self::TYPE_MTP;
            }
            // issue #5648
            /*
             if( $this->hasKey( 'g729_channels' ) )
             {
             $this->type |= self::TYPE_G729;
             }
             */
            if( $this->hasKey( 'voip_channels' ) )
            {
                $this->type |= self::TYPE_VOIP;
            }
            if( $this->hasKey( 'gateway_enable' ) )
            {
                $this->type |= self::TYPE_GATEWAYENABLE;
            }
            if( $this->hasKey( 'gateway_links' ) )
            {
                $this->type |= self::TYPE_GATEWAYLINKS;
            }
            if( $this->hasKey( 'gateway_gsm_channels' ) )
            {
                $this->type |= self::TYPE_GATEWAYGSMCHANNELS;
            }
            if( $this->hasKey( 'play_channels' ) )
            {
                $this->type |= self::TYPE_PLAYCHANNELS;
            }
            if( $this->hasKey( 'record_channels' ) )
            {
                $this->type |= self::TYPE_RECORDCHANNELS;
            }
            if( $this->hasKey( 'gateway_analytics' ) )
            {
                $this->type |= self::TYPE_GATEWAYANALYTICS;
            }
            if( $this->hasKey( 'gateway_analytics_trunk0' ) )
            {
                $this->type |= self::TYPE_GATEWAYANALYTICSTRUNK0;
            }
            if( $this->hasKey( 'gateway_analytics_trunk1' ) )
            {
                $this->type |= self::TYPE_GATEWAYANALYTICSTRUNK1;
            }
            if( $this->hasKey( 'gateway_analytics_trunk2' ) )
            {
                $this->type |= self::TYPE_GATEWAYANALYTICSTRUNK2;
            }
            if( $this->hasKey( 'gateway_analytics_trunk3' ) )
            {
                $this->type |= self::TYPE_GATEWAYANALYTICSTRUNK3;
            }
            if( $this->hasKey( 'gateway_analytics_voip_calls' ) )
            {
                $this->type |= self::TYPE_GATEWAYANALYTICSVOIPCALLS;
            }
            if( $this->hasKey( 'gateway_voip_calls' ) )
            {
                $this->type |= self::TYPE_GATEWAYVOIPCALLS;
            }
            // easy to use issued_date and expiration_date objects
            $this->issued = explode( '-', $this->issued_date );
            $this->expires = explode( '-', $this->expiration_date );
        }

        /**
         * Return licenses filename with full path
         *
         * @param manager LicenseManager The license manager
         * @param name string File name
         * @param testifexits boolean Test if file exists
         * @return string
         */
        public static function getFilenameWithFullPath( $manager, $name, $testifexits = true )
        {
            if( is_null( $manager ) )
            {
                throw new KLicenseException( 'license-error-not-found', self::EIO );
            }
            if( is_null( $name ) )
            {
                throw new KLicenseException( 'license-error-not-found', self::EIO );
            }
            if( PHP_OS == 'WINNT' )
            {
                $_filename = "{$manager->getLicensesDirectory()}\\{$name}";
            }
            else
            {
                $_filename = "{$manager->getLicensesDirectory()}/{$name}";
            }
            if( $testifexits && !is_file( $_filename ) )
            {
                throw new KLicenseException( 'license-error-not-found', self::EIO );
            }
            return $_filename;
        }

        /**
         * Get issued date as array
         *
         * @return array
         */
        public function getIssued( )
        {
            return $this->issued;
        }

        /**
         * Get expires date as array
         *
         * @return array
         */
        public function getExpires( )
        {
            return $this->expires;
        }

        /**
         * Check if expired
         *
         * @return boolean
         */
        public function isExpired( )
        {
            return $this->checkIfExpiresIn( $this->manager->getNow( ) );
        }

        /**
         * Check if will expire soon
         *
         * @return boolean
         */
        public function willExpire( )
        {
            return $this->checkIfExpiresIn( $this->manager->getWillExpire( ) );
        }

        /**
         * Check if the license will be expired in supplied date
         *
         * @param date string date to be compared to
         * @return boolean
         */
        public function checkIfExpiresIn( $date )
        {
            $parts = explode( '-', $date );
            $reftime = (int)"{$parts[0]}{$parts[1]}{$parts[2]}";
            $expirestime = (int)"{$this->expires[0]}{$this->expires[1]}{$this->expires[2]}";
            return $expirestime < $reftime;
        }

        /**
         * Get the file name
         *
         * @return string
         */
        public function getFilename( )
        {
            return $this->filename;
        }

        /**
         * Magic method to read license values as $obj->keyname
         *
         * @param string name License key
         * @return string
         */
        public function __get( $name )
        {
            return $this->get( $name );
        }

        /**
         * Get license key value
         *
         * @param string name Key
         * @return string
         */
        public function get( $name )
        {
            if( $this->hasKey( $name ) )
            {
                return $this->license[$name];
            }
            throw new KLicenseException( 'license-error-invalid-key', self::EINVALID );
        }

        /**
         * Check if license has key
         *
         * @param string name Key
         * @return boolean
         */
        public function hasKey( $name )
        {
            return array_key_exists( $name, $this->license );
        }

        /**
         * Get existing license keys
         *
         * @return array
         */
        public function getKeys( )
        {
            return array_keys( $this->license );
        }

        /**
         * Get type name
         *
         * The returned value is a suitable value to KLicense::get method
         *
         * @param int type
         * @return string Type name
         */
        public function getTypeAttributeName( $type )
        {
            switch( $type )
            {
                case self::TYPE_FAX:
                    return 'fax_channels';
                case self::TYPE_ISUP:
                    return 'isup_circuits';
                case self::TYPE_MTP2:
                    return 'mtp2_links';
                case self::TYPE_MTP:
                    return 'mtp_dual_mode';
                // issue 5648
                /*
                 case self::TYPE_G729:
                 return 'g729_channels';
                 */
                case self::TYPE_VOIP:
                    return 'voip_channels';
                case self::TYPE_GATEWAYENABLE:
                    return 'gateway_enable';
                case self::TYPE_GATEWAYLINKS:
                    return 'gateway_links';
                case self::TYPE_GATEWAYGSMCHANNELS:
                    return 'gateway_gsm_channels';
                case self::TYPE_PLAYCHANNELS:
                    return 'play_channels';
                case self::TYPE_RECORDCHANNELS:
                    return 'record_channels';
                case self::TYPE_GATEWAYANALYTICS:
                    return 'gateway_analytics';
                case self::TYPE_GATEWAYANALYTICSTRUNK0:
                    return 'gateway_analytics_trunk0';
                case self::TYPE_GATEWAYANALYTICSTRUNK1:
                    return 'gateway_analytics_trunk1';
                case self::TYPE_GATEWAYANALYTICSTRUNK2:
                    return 'gateway_analytics_trunk2';
                case self::TYPE_GATEWAYANALYTICSTRUNK3:
                    return 'gateway_analytics_trunk3';
                case self::TYPE_GATEWAYANALYTICSVOIPCALLS:
                    return 'gateway_analytics_voip_calls';
                case self::TYPE_GATEWAYVOIPCALLS:
                    return 'gateway_voip_calls';
                default:
                    throw new KLicenseException( 'license-error-invalid-key', self::EINVALID );
            }
        }

        /**
         * Get type as string
         *
         * Formatting options:
         *   %n - Type name
         *   %v - Type key value
         *
         * @param glue string types glue
         * @param format string each element formatting settings
         * @return string
         */
        public function getTypeAsString( $glue = ', ', $fmt = '%n' )
        {
            $types = array( );
            if( $this->hasType( self::TYPE_FAX ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'fax', $this->fax_channels );
            }
            if( $this->hasType( self::TYPE_ISUP ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'isup', $this->isup_circuits );
            }
            if( $this->hasType( self::TYPE_MTP2 ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'mtp2', $this->mtp2_links );
            }
            if( $this->hasType( self::TYPE_MTP ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'mtp', $this->mtp_dual_mode );
            }
            // issue #5648
            /*
            if( $this->hasType( self::TYPE_G729 ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'g729', $this->g729_channels );
            }
            */
            if( $this->hasType( self::TYPE_VOIP ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'voip', $this->voip_channels );
            }
            if( $this->hasType( self::TYPE_GATEWAYENABLE ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'gateway', $this->gateway_enable );
            }
            if( $this->hasType( self::TYPE_GATEWAYLINKS ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'e1', $this->gateway_links );
            }
            if( $this->hasType( self::TYPE_GATEWAYGSMCHANNELS ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'gsm', $this->gateway_gsm_channels );
            }
            if( $this->hasType( self::TYPE_PLAYCHANNELS ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'play', $this->play_channels );
            }
            if( $this->hasType( self::TYPE_RECORDCHANNELS ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'record', $this->record_channels );
            }
            if( $this->hasType( self::TYPE_GATEWAYANALYTICS ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'analytics', $this->gateway_analytics );
            }
            if( $this->hasType( self::TYPE_GATEWAYANALYTICSTRUNK0 ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'analytics_trunk0', $this->gateway_analytics_trunk0 );
            }
            if( $this->hasType( self::TYPE_GATEWAYANALYTICSTRUNK1 ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'analytics_trunk1', $this->gateway_analytics_trunk1 );
            }
            if( $this->hasType( self::TYPE_GATEWAYANALYTICSTRUNK2 ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'analytics_trunk2', $this->gateway_analytics_trunk2 );
            }
            if( $this->hasType( self::TYPE_GATEWAYANALYTICSTRUNK3 ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'analytics_trunk3', $this->gateway_analytics_trunk3 );
            }
            if( $this->hasType( self::TYPE_GATEWAYANALYTICSVOIPCALLS ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'analytics_voip_calls', $this->gateway_analytics_voip_calls );
            }
            if( $this->hasType( self::TYPE_GATEWAYVOIPCALLS ) )
            {
                $types[] = $this->formatTypeAsString( $fmt, 'voip_calls', $this->gateway_voip_calls );
            }
            return implode( $glue, $types );
        }

        /**
         * Format type as string
         *
         * @param format string
         * @param name string type name as string
         * @param value string type key value
         * @return string
         */
        protected function formatTypeAsString( $fmt, $name, $value )
        {
            return str_replace( '%v', $value, str_replace( '%n', $name, $fmt ) );
        }

        /**
         * Check if the license is has 'needle' type
         *
         * @param int needle
         * @return boolean
         */
        public function hasType( $needle )
        {
            return ( ( $this->type & $needle ) == $needle );
        }

        /**
         * Check if the license is has any type
         *
         * @param int any
         * @return boolean
         */
        public function hasAnyType( $any )
        {
            return ( ($this->type & $any) != 0 );
        }

        /**
         * Check if this license has signature block
         *
         * @return boolean
         */
        public function hasSignature( )
        {
            return $this->signatureFound;
        }

        /**
         * Check if the license is associated to some device
         *
         * @return boolean
         */
        public function hasConfiguredDevice() {

            require_once KCONFIG . '/classes/KDevicesYaml.php';

            $kyaml = new KDevicesYaml( );
            $yaml = $kyaml->getYaml( );

            if( isset( $yaml['Devices'] ) === false )
            {
                return false;
            }

            foreach( $yaml['Devices'] as &$device )
            {
                if ( array_key_exists( $this->serial_number, $device ) )
                {
                    return true;
                }
            }

            return false;
        }

        /**
         * Public method to get a formatted serial number if hardware type is HARDWARE_ID
         *
         * @return boolean
         */
        public function formatted_serial_number()
        {
            if ( $this->hardware_type === 'HARDWARE_ID' )
            {
                return substr( $this->serial_number, 0, 16 ) . "...";
            }
            
            return $this->serial_number;
        }

        /**
         * Public method to check if the license is valid
         *
         * @return boolean
         */
        public function isValid()
        {
            if ( $this->hardware_type === 'HARDWARE_ID' )
            {
                $hardware_id = "";

                if( !isset( $_SESSION['gateway-hardware_id'] ) )
                {
                    exec( "/usr/sbin/kwebhelper gethardwareid", $output, $excode );
                    if( $excode == 0 )
                    {
                        $hardware_id = $output[0];
                    }
                    $_SESSION['gateway-hardware_id'] = $hardware_id;
                }
                else
                {
                    $hardware_id = $_SESSION['gateway-hardware_id'];
                }

                return ( $this->serial_number === $hardware_id );
            }

            if ( $this->hardware_type === 'HASP_HARDLOCK' )
            {
                return true;
            }
            
            if ( $this->hasConfiguredDevice() )
            {
                if ( !$this->isExpired() )
                {
                    return true;
                }
            }
            
            return false;
        }
    }
