<?php
    /**
     * Application base class
     */
    final class KSystem
    {
        /**
         * Current language
         */
        private $language = null;
        /**
         * Current version
         */
        private $version;
        /**
         * Self reference
         */
        static private $self = null;
        /**
         * Challenge _SESSION key
         */
        const CHALLENGE_SESSION_KEY = 'auth::challenge';
        /**
         * Challenge-response user
         */
        const CHALLENGE_RESPONSE_USER = 'challenge-response';

        /**
         * WARNING: Singleton implementation!
         */
        private function __construct( )
        {
            @session_start( );
            if( !IS_GATEWAY )
            {
                if( isset( $_SESSION['time_zone'] ) )
                    date_default_timezone_set( $_SESSION['time_zone'] );
                else
                    date_default_timezone_set( 'Brazil/East' );
            }

            if( array_key_exists( 'language', $_GET ) )
            {
                $this->setLanguage( $_GET['language'] );
            }
            elseif( array_key_exists( 'language', $_SESSION ) )
            {
                $this->setLanguage( $_SESSION['language'] );
            }
            else
            {
                $this->detectLanguage( );
            }
        }

        static public function getInstance( )
        {
            if( is_null( self::$self ) )
            {
                self::$self = new KSystem( );
            }
            return self::$self;
        }

        /**
         * Translate word
         */
        public function translate( $word )
        {
            //TODO: Change the dictionary to be loaded by the setLanguage
            // function
            //      as an protected attribute of this class
            global $dicionario;

            if( isset( $dicionario[$this->getLanguage( )][$word] ) )
            {
                return $dicionario[$this->getLanguage( )][$word];
            }
            else
            {
                trigger_error( "Undefined translation key {$word} ({$this->getLanguage()})" );
                return $word;
            }
        }

        /**
         * Detect and set language from browser headers or default to pt-br
         */
        public function detectLanguage( )
        {
            $this->setLanguage( 'pt-br' );
        }

        public function setLanguage( $lang )
        {
            $this->language = $lang;
            $_SESSION['language'] = $lang;
        }

        public function getLanguage( )
        {
            return $this->language;
        }

        /**
         * Get the root configuration directory
         */
        public function getConfigDir( )
        {
            if( PHP_OS == 'WINNT' )
            {
                if( !isset( $_SESSION['configPath'] ) )
                {
                    $winVersion = PHP_WINDOWS_VERSION_MAJOR . "." . PHP_WINDOWS_VERSION_MINOR;
                    exec( KROOT . '\\..\\..\\Tools\\kwebhelper configdir', $sysdir );
                    if( $winVersion >= 6.1 )// WIN7+
                    {
                        $sysdata = explode( "\\", $sysdir[0] );
                        $path = sizeof( $sysdata ) <= 3  ? "{$sysdata[0]}\\ProgramData\\{$sysdata[1]}\\{$sysdata[2]}" : $sysdir[0];
                    }
                    else
                         $path = $sysdir[0];

                    $_SESSION['configPath'] = $path;
                }
                return $_SESSION['configPath'];
            }
            else
            {
                return "/etc/khomp/config";
            }
        }

        /**
         * Get the installation directory
         */
        public function getInstallDir( )
        {
            if( PHP_OS == 'WINNT' )
            {
                $tmpDir = explode( "\\", $this->getConfigDir( ) );
                array_pop( $tmpDir );
                return implode( "\\", $tmpDir );
            }
            else
            {
                return '/etc/khomp';
            }
        }

        /**
         * Get the package version dir
         */
        public function getPackageVersionDir( )
        {
            $install = $this->getInstallDir( );

            if( !is_null( $install ) )
            {
                if( PHP_OS == 'WINNT' )
                    $dir = "{$install}\\data";
                else
                    $dir = "{$install}/data";

                return $dir;
            }

            return '';
        }

        /**
         * Get the package version
         */
        public function getPackageVersion( )
        {
            $dir = $this->getPackageVersionDir( );

            if ( $dir !== '' )
            {
                if( PHP_OS == 'WINNT' )
                    $file = "{$dir}\\k3l.dat";
                else
                    $file = "{$dir}/k3l.dat";

                if( file_exists( $file ) )
                {
                    $lines = file( $file );
                    foreach( $lines as $line )
                    {
                        $line = trim( $line );
                        if( strpos( $line, 'LastBuildInstalled' ) === 0 )
                        {
                            $start = strpos( $line, '=' ) + 1;
                            return substr( $line, $start, strlen( $line ) - $start );
                        }
                    }
                }
            }

            return $this->translate( 'InfoNotAvailable' );
        }

        /**
         * Redirect to login page, if the user is not logged in
         */
        public function redirectIfNotLogged( )
        {
            if( !array_key_exists( 'login', $_SESSION ) )
            {
                if( isset( $_REQUEST['rawmode'] ) )
                    header( 'HTTP/1.1 401 Unauthorized' );
                else
                    header( 'location: /?error=Kerr3' );
                exit ;
            }
        }

        /**
         * Check access permission
         *
         * @param location string System location
         * @param type string|array Required permission type or permission types
         * @return boolean true if has permission (if type is string) or any
         * permission
         *    (if type is array)
         */
        public function hasPermission( $location, $type = 'full' )
        {
            if( array_key_exists( 'loginAccess', $_SESSION ) && array_key_exists( $location, $_SESSION['loginAccess'] ) )
            {
                if( is_array( $type ) )
                {
                    return in_array( $_SESSION['loginAccess'][$location], $type );
                }
                else
                {
                    return $_SESSION['loginAccess'][$location] == $type;
                }
            }
            return false;
        }

        /**
         * Slow and possibly unsafe random string
         *
         * @param length integer Number of bytes
         * @return string
         */
        protected function authenticationRandomBytes( $length )
        {
            $str = '';
            for( $i = 0; $i < $length; ++$i )
            {
                $str .= chr( rand( 0, 255 ) );
            }
            return $str;
        }

        /**
         * True if is authenticating via challenge-response, false otherwise
         *
         * @return bool
         */
        public function authenticationUsingChallengeResponse( )
        {
            return isset( $_SESSION[self::CHALLENGE_SESSION_KEY] );
        }

        /**
         * Create challenge
         *
         * @return string
         */
        public function authenticationChallenge( )
        {
            if( PHP_OS == 'WINNT' )
            {
                $bytes = $this->authenticationRandomBytes( 10240 );
            }
            else
            {
                $devurandom = fopen( '/dev/urandom', 'r' );
                if( !$devurandom )
                {
                    $bytes = $this->authenticationRandomBytes( 10240 );
                }
                else
                {
                    $bytes = fread( $devurandom, 10240 );
                    fclose( $devurandom );
                }
            }
            $_SESSION[self::CHALLENGE_SESSION_KEY] = substr( base64_encode( sha1( $bytes, true ) ), 0, 10 );

            return $_SESSION[self::CHALLENGE_SESSION_KEY];
        }

        /**
         * Check if the response to the challenge is correct
         *
         * @param response string Response to be checked
         * @return bool
         */
        public function authenticationResponse( $response )
        {
            if( !$this->authenticationUsingChallengeResponse() )
            {
                return false;
            }
            $_response = substr( base64_encode( sha1( '+KWebPortal+' . $_SESSION[self::CHALLENGE_SESSION_KEY], true ) ), 0, 10 );
            $_SESSION[self::CHALLENGE_SESSION_KEY] = self::CHALLENGE_RESPONSE_USER;
            return $response === $_response;
        }

        public function isCTIorSPX( )
        {
            $package = strtolower( $this->getPackageVersion( ) );

            if( strpos( $package, 'k3l' ) === 0 )
                return 'CTI';
            else if( strpos( $package, 'channel' ) === 0 || strpos( $package, 'mod' ) === 0 )
                return 'SPX';
            else
                return '';
        }

        public function translateErrMessage( $str )
        {
            $str = trim( $str );
            if( strpos( ',', $str ) >= 0 )
            {
                $arr = explode( ',', $str );
                $resp = Array( );
                foreach( $arr as $k => $v )
                {
                    if( strpos( trim( $v ), 'Query failed' ) !== false )
                        $v = str_replace( 'Query failed', $this->translate( 'query-failed' ), $v );
                    elseif( strpos( trim( $v ), 'Command failed' ) !== false )
                        $v = str_replace( 'Query failed', $this->translate( 'command-failed' ), $v );
                    else
                        $v = $this->translate( $v );
                    $resp[] = $v;
                }

                return implode( ' - ', $resp );
            }
            else
                return $this->translate( $str );
        }

        public function getTimeZone()
        {
            return date_default_timezone_get();
        }

        public function setTimeZone($timezone_id)
        {
            $_SESSION['time_zone'] = $timezone_id;
            date_default_timezone_set($timezone_id);
        }
    }
