<?php
    require_once KCONFIG . '/classes/classKYaml.php';

    class KFirewallYamlException extends KYamlException
    {
        /* VOID */
    }

    /**
     * Firewall YAML definition file.
     * 
     * WARNING: This file manipulation *requires* two phase write sequence,
     * each of them being executed by different HTTP requestes, to avoid user
     * lockout by adding a bad firewall rule that lock's him outside the
     * gateway.
     * 
     * First: The file is staged as kfirewall-stage.yaml, so kfirewall can load
     *        the staged configuration to netfilter apply it.
     * Second: The staged file is renamed to kfirewall.yaml.
     * 
     * If the second phase is reached and executed without errors, this means
     * that the user has no lockout rule defined. The rules load in the first
     * phase _should_ leave a background timer based process to remove staged
     * configuration and revert to (reload) the previous rules. This background
     * process _should_ be killed by the second phase (if any).
     * 
     * The second phase may be reached after timed revert and, in this case,
     * _should_ fail.
     */
    class KFirewallYaml extends KYaml
    {
        public static $protocols = array( 'ANY', 'IP', 'TCP', 'UDP', 'ICMP' );
        public static $actions = array( 'ACCEPT', 'REJECT', 'DROP', /* 'CHAIN', */ 'RETURN' );
        public static $states = array( 'NEW', 'RELATED', 'ESTABLISHED', 'INVALID' );
        public static $trackingModes = array( 'NO_TRACK', 'CHECK', 'MARK' );

        protected $stagefile;

        const ROLLBACK_SIGNAL = '/tmp/kfirewall-rollback';

        public function __construct( )
        {
            //TODO: rules!

            // lazy loaded
            unset( $this->stagefile );

            $app = KSystem::getInstance( );
            $this->validationProgram = '/usr/sbin/kfirewall --validate-config';
            parent::__construct( $app->getConfigDir( ) . DIRECTORY_SEPARATOR . 'kfirewall.yaml' );
        }

        public function __get( $name )
        {
            switch( $name )
            {
                case 'stagefile':
                    $app = KSystem::getInstance();
                    return $app->getConfigDir( ) . DIRECTORY_SEPARATOR . 'kfirewall-stage.yaml';
                    break;
            }
            trigger_error( "Invalid attribute {$name}", E_USER_WARNING );
            return null;
        }

        ////////////////////////////////////////////////////////////////////////
        // Stage and save
        ////////////////////////////////////////////////////////////////////////
        /**
         * Stage the configuration file
         * This function will write the current configuration to file name
         * kfirewall-stage.yaml instead kfirewall.yaml
         */
        public function stage( )
        {
            if( file_exists( $this->stagefile ) )
            {
                throw new KFirewallYamlException( 'kfirewallyaml-error-already-staged' );
            }

            $this->setting_dump_force_quotes = true;
            $this->validate( $this->rules, $this->yaml, true );

            if( file_put_contents( $this->stagefile, $this->getHeader( ) . $this->dump( $this->yaml ), LOCK_EX ) === false )
            {
                throw new KFirewallYamlException( 'kfirewallyaml-error-write-failed' );
            }

            exec( "/etc/init.d/kfirewall stage 2>&1", $output, $errcode );
            if( $errcode != 0 )
            {
                // stage error
                unlink( $this->stagefile );
                throw new KFirewallYamlException( implode( ' ', $output ) );
            }
        }

        /**
         * Rename staged file to consolidated file
         */
        public function save( )
        {
            if( file_exists( self::ROLLBACK_SIGNAL ) )
            {
                unlink( self::ROLLBACK_SIGNAL );
                throw new KFirewallYamlException( 'kfirewallyaml-error-commit-timeout-configuration-rollback' );
            }

            if( !file_exists( $this->stagefile ) )
            {
                throw new KFirewallYamlException( 'kfirewallyaml-error-need-stage-first' );
            }

            $this->backup( );
            exec( "/etc/init.d/kfirewall commit 2>&1", $output, $errcode );
            if( $errcode != 0 )
            {
                $this->rollback( );
                //throw new KFirewallYamlException( implode( ' ', $output ) );
                throw new KFirewallYamlException( 'kfirewallyaml-error-write-failed' );
            }
            $this->commit( );
        }

        ////////////////////////////////////////////////////////////////////////
        // Generic functions
        ////////////////////////////////////////////////////////////////////////

        /**
         * Get all chains from the supplied table
         * 
         * @param string $table Table name
         * @param bool $corechains If true return core chains if not return user chains
         * @return array
         */
        protected function _getAllChains( $table, $corechains )
        {
            $chainset = $corechains ? 'Core' : 'Chains';

            return isset( $this->yaml[$table][$chainset] ) ? $this->yaml[$table][$chainset] : array(); 
        }

        /**
         * Get chain from supplied table
         * 
         * @param string $table Table name
         * @param string $name Chain name
         * @param bool $corechain If true, search in core chains if not search in user chains
         * @return array
         */
        protected function _getChain( $table, $name, $corechain )
        {
            $chainset = $corechain ? 'Core' : 'Chains';

            if( isset( $this->yaml[$table][$chainset] ) )
            {
                foreach( $this->yaml[$table][$chainset] as $chain )
                {
                    if( $chain['Name'] === $name )
                    {
                        return $chain;
                    }
                }
            }
            return array();
        }

        /**
         * Set chain to supplied table
         * 
         * @param string $table Table name
         * @param array $data Chain data
         * @param string $oldName Chain old name
         */
        protected function _setChain( $table, $data, $oldName )
        {
            if( !isset( $this->yaml[$table]['Chains'] ) )
            {
                $this->yaml[$table]['Chains'] = array();
            }

            //TODO: validate,rename

            foreach( $this->yaml[$table]['Chains'] as &$chain )
            {
                if( $chain['Name'] === $data['Name'] )
                {
                    $chain = $data;
                    return;
                }
            }
            $this->yaml[$table]['Chains'][] = $data;
        }

        ////////////////////////////////////////////////////////////////////////
        // Filter functions
        ////////////////////////////////////////////////////////////////////////

        public function getAllFilterChains( $core = false )
        {
            return $this->_getAllChains( 'Filter', $core );
        }

        public function getFilterChain( $name, $core = false )
        {
            return $this->_getChain( 'Filter', $name, $core );
        }

        public function setFilterChain( $data, $oldName )
        {
            return $this->_setChain( 'Filter', $data, $oldName );
        }

        public function getFilterChainRule( $name, $chain, $core = false )
        {
            $chain_ = $this->getFilterChain( $chain, $core );
            foreach( $chain_['Rules'] as &$rule )
            {
                if( $rule['Chain'] === $name )
                {
                    return $rule;
                }
            }
            return array();
        }

    }
