<?php
    class KGatewayProfileAudioFileManagerException extends Exception
    {
        /* VOID */
    };

    class KGatewayProfileAudioFileManager
    {
        protected $rootdir = '/var/local/ftp/audio';
        protected $cache = array( );
        //@formatter:off
        // from KWaveHaeder @ KWave.h
        protected $supportedFormats = array(
            /*PCM*/ 1,
            /*ALAW*/ 6,
            /*ULAW*/ 7,
            /*DVI*/ 17,
            /*GSM*/ 49
        );
        protected $formatString = array(
            1 => 'PCM',
            6 => 'ALAW',
            7 => 'ULAW',
            17 => 'DVI',
            49 => 'GSM'
        );
        //@formatter:on

        public function __construct( )
        {
            /* VOID */
        }

        /**
         * Get valid files
         *
         * @return array
         */
        public function getFiles( )
        {
            if( count( $this->cache ) === 0 )
            {
                $files = glob( "{$this->rootdir}/*", GLOB_NOSORT );

                if( $files === false )
                {
                    throw new KGatewayProfileAudioFileManagerException( 'gateway-profile-error-unable-to-read-audio-directory' );
                }
                elseif( $files )
                {
                    natcasesort( $files );
                    foreach( $files as $f )
                    {
                        $f = basename( $f );
                        // skip files that start with a dot and directories
                        if( substr( $f, 0, 1 ) === '.' || is_dir( "{$this->rootdir}/${f}" ) )
                        {
                            continue;
                        }
                        $this->cache[$f] = $this->getFileDecription( $f );
                    }
                }
            }
            return $this->cache;
        }

        /**
         * Read header and compose file description
         *
         * @see KWaveHeader @ KWave.h
         * @param string $file file name
         * @return array( boolean, string ) true if is valid and file description
         */
        protected function getFileDecription( $file )
        {
            $filename = "{$this->rootdir}/{$file}";
            $st = stat( $filename );
            // stat failed or file is too small
            if( $st === false || $st['size'] < 44 )
            {
                $sz = $st ? $st['size'] : 0;
                trigger_error( "{$file} stat(2) failed or file size is too short, expected at least 44 bytes long, found {$sz}", E_USER_NOTICE );
                if( $st === false )
                {
                    return array( false, "{$file} (invalid file, unable to query file properties)" );
                }
                else
                {
                    return array( false, "{$file} (invalid file, too short, {$this->toHumanSize( $st['size'] )})" );
                }
            }

            $_f = fopen( $filename, 'r' );
            if( $_f === false )
            {
                trigger_error( "{$file} fopen(2) failed", E_USER_NOTICE );
                return array( false, "{$file} (invalid file, unable to open file, {$this->toHumanSize( $st['size'] )})" );
            }
            $header = fread( $_f, 44 );
            fclose( $_f );

            // 0-3 always RIFF
            if( substr( $header, 0, 4 ) != 'RIFF' )
            {
                trigger_error( "{$file} is not in RIFF format", E_USER_NOTICE );
                return array( true, "{$file} (ALAW raw file, {$this->toHumanSize( $st['size'] )})" );
            }

            // 4-7 file size
            $data = unpack( 'Vsize', substr( $header, 4, 4 ) );
            if( ($st['size'] - 8) != $data['size'] )
            {
                trigger_error( "{$file} file size missmatch, expected {$st['size']} - 8, found {$data['size']}", E_USER_NOTICE );
                return array( false, "{$file} (invalid RIFF file, file size did not match, {$this->toHumanSize( $st['size'] )})" );
            }

            // 8-11 always WAVE
            // 12-15 always "fmt "
            if( substr( $header, 8, 8 ) != 'WAVEfmt ' )
            {
                trigger_error( "{$file} RIFF file is not a WAVE file", E_USER_NOTICE );
                return array( false, "{$file} (invalid RIFF file, expected WAVE format, {$this->toHumanSize( $st['size'] )})" );
            }

            // 16-19 is the length of the "fmt " header chunk
            // 20-21 audio format
            // 22-23 number of channels
            // 24-27 sample rate
            // 28-31 byte rate
            // 32-33 block align
            // 34-35 Bits per sample
            $data = unpack( 'Vchunksize/vformat/vchannels/Vsamplerate/Vbyterate/vblockalign/vbitspersample', substr( $header, 16, 20 ) );
            if( $data['chunksize'] < 16 || $data['chunksize'] > 100 )
            {
                trigger_error( "{$file} invalid chunk size {$data['chunksize']}", E_USER_NOTICE );
                return array( false, "{$file} (invalid RIFF file, fmt size out of bounds, {$this->toHumanSize( $st['size'] )})" );
            }
            if( !in_array( $data['format'], $this->supportedFormats ) )
            {
                trigger_error( "{$file} audio format {$data['format']} is not supported", E_USER_NOTICE );
                return array( false, "{$file} (unsupported format, {$this->toHumanSize( $st['size'] )})" );
            }
            $fmt = $this->formatString[$data['format']];
            if( $data['channels'] != 1 )
            {
                trigger_error( "{$file} invalid number of channels, expected 1, found {$data['channels']}", E_USER_NOTICE );
                return array( false, "{$file} ({$fmt}, unsupported number of channels {$data['channels']}, {$this->toHumanSize( $st['size'] )})" );
            }
            $spec = array( true, "{$file} ($fmt, {$data['bitspersample']} bits, {$data['samplerate']} hz, {$this->toHumanSize( $st['size'] )})" );
            switch( $data['format'] )
            {
                // PCM
                case 1:
                    if( $data['samplerate'] == 8000 )
                    {
                        if( $data['bitspersample'] == 8 || $data['bitspersample'] == 16 )
                        {
                            return $spec;
                        }
                    }
                    elseif( $data['samplerate'] > 11000 && $data['samplerate'] < 11050 && $data['bitspersample'] == 8 )
                    {
                        return $spec;
                    }
                    break;
                // ALAW, ULAW
                case 6:
                case 7:
                    if( $data['samplerate'] == 8000 && $data['bitspersample'] == 8 )
                    {
                        return $spec;
                    }
                    break;
                // DVI
                case 17:
                    if( $data['samplerate'] == 8000 && $data['bitspersample'] == 8 && $data['blockalign'] <= 512 )
                    {
                        return $spec;
                    }
                    break;
                // GSM
                case 49:
                    if( $data['samplerate'] == 8000 && $data['blockalign'] == 65 && $data['byterate'] == 1625 )
                    {
                        return $spec;
                    }
                    break;
            }

            trigger_error( "{$file} invalid parameters for format {$data['format']}", E_USER_NOTICE );
            return array( false, "{$file} ({$fmt}, unsupported mode, {$this->toHumanSize( $st['size'] )})" );
        }

        /**
         * Format size to human readable
         * 
         * @param int $size
         * @return string
         */
        protected function toHumanSize( $size )
        {
            if( $size > 1048576 )
            {
                $unit = 'MiB';
                $size = $size / 1048576;
            }
            elseif( $size > 1024 )
            {
                $unit = 'KiB';
                $size = $size / 1024;
            }
            else
            {
                $unit = 'B';
            }

            $size = round( $size, 2 );
            return "{$size} {$unit}";
        }
    }
