<?php

    class KGatewayNTPException extends Exception
    {
    }

    class KGatewayNTP
    {
        /**
         * array Configuration data
         */
        protected $configuration = array( );
        /**
         * string Configuration file.
         */
        const CONFIG = '/etc/khomp/config/kgateway-ntp.cfg';

        /**
         * int Input/Output error.
         */
        const EIO = 1;
        /**
         * int Runtime error.
         */
        const ERUNTIME = 2;
        /**
         * int Invalid parameter/value
         */
        const EINVALID = 3;

        // @formatter:off
        /**
         * array Validation stuff
         * 
         * Format: key => array
         *           verification method => array|string
         * 
         * The verification method is a callback name to any of this class
         * validators, if it is an array, then the first element of the array
         * is the callback name and the second an array with that will be
         * supplied to the callback as second parameter to the function call.
         */
        static protected $stanzas = array(
            'ntp-enabled' => 'checkYesNo',
            'ntp-timeout' => array( 'checkInteger', array( 'min' => 1, 'max' => 999 ) ),
            'ntp-servers' => 'checkFqdnList'
        );
        // @formatter:on

        public function __construct( )
        {
            /* VOID */
        }

        /**
         * Parse data
         */
        public function load( )
        {
            if( !file_exists( self::CONFIG ) )
            {
                throw new KGatewayNTPException( 'gateway-ntp-error-configuration-not-found', self::EIO );
            }
            $lines = file( self::CONFIG, FILE_IGNORE_NEW_LINES );
            if( empty( $lines ) )
            {
                throw new KGatewayNTPException( 'gateway-ntp-error-configuration-read-failed', self::EIO );
            }

            foreach( $lines as $line )
            {
                $parts = explode( '=', $line );
                $this->configuration[str_replace( '_', '-', strtolower( $parts[0] ) )] = str_replace( '"', '', $parts[1] );
            }
        }

        /**
         * Save configuration
         */
        public function save( )
        {
            if( count( $this->configuration ) == 0 )
            {
                throw new KGatewayNTPException( 'gateway-ntp-error-invalid-config', self::EINVALID );
            }

            $buffer = '';
            foreach( $this->configuration as $key => $value )
            {
                $buffer .= str_replace( '-', '_', strtoupper( $key ) ) . "=\"{$value}\"\n";
            }

            $_tmpname = self::CONFIG . '.in';
            if( file_put_contents( $_tmpname, $buffer ) === false )
            {
                throw new KGatewayNTPException( 'gateway-ntp-error-write-failed', self::EIO );
            }
            if( rename( $_tmpname, self::CONFIG ) === false )
            {
                unlink( $_tmpname );
                throw new KGatewayNTPException( 'gateway-ntp-error-write-failed', self::EIO );
            }
        }

        /**
         * Get the configuration array
         */
        public function getConfiguration( )
        {
            return $this->configuration;
        }

        /**
         * Set the configuration array
         */
        public function setConfiguration( $config )
        {
            if( !is_array( $config ) )
            {
                throw new KGatewayNTPException( 'gateway-ntp-error-invalid-config', self::EINVALID );
            }
            foreach( $config as $key => &$value )
            {
                if( array_key_exists( $key, self::$stanzas ) )
                {
                    $validationMethod = null;
                    $validationData = null;
                    if( is_array( self::$stanzas[$key] ) )
                    {
                        $validationMethod = self::$stanzas[$key][0];
                        $validationData = self::$stanzas[$key][1];
                    }
                    else
                    {
                        $validationMethod = self::$stanzas[$key];
                    }
                    if( method_exists( $this, $validationMethod ) )
                    {
                        if( is_null( $validationData ) )
                        {
                            $ok = $this->$validationMethod( $value );
                        }
                        else
                        {
                            $ok = $this->$validationMethod( $value, $validationData );
                        }
                        if( !$ok )
                        {
                            throw new KGatewayNTPException( 'gateway-ntp-error-invalid-value', self::EINVALID );
                        }
                    }
                    else
                    {
                        throw new KGatewayNTPException( 'gateway-ntp-error-validator-not-found', self::ERUNTIME );
                    }
                }
                else
                {
                    throw new KGatewayNTPException( 'gateway-ntp-error-invalid-stanza', self::EINVALID );
                }
            }

            // stanzas seems ok, now validate if all stanzas are set
            $missingstanzas = array_diff_key( self::$stanzas, $config );
            if( count( $missingstanzas ) > 0 )
            {
                throw new KGatewayNTPException( 'gateway-ntp-error-invalid-config', self::EINVALID );
            }

            $this->configuration = $config;
        }

        // ********************** validation methods **************************
        public function checkFqdn( &$value, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $value ) )
            {
                return true;
            }
            return preg_match( '/([a-zA-Z0-9\-_]+\.)+[a-zA-Z0-9]{2,3}/', $value ) === 1;
        }

        public function checkFqdnList( &$list, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $value ) )
            {
                return true;
            }
            if( is_array( $list ) )
            {
                $l = &$list;
            }
            else
            {
                $l = explode( ' ', $list );
            }
            if( count( $l ) == 0 )
            {
                return false;
            }
            foreach( $l as &$v )
            {
                if( !$this->checkFqdn( $v, $data ) )
                {
                    return false;
                }
            }
            if( is_array( $list ) )
            {
                $_l = implode( ' ', $list );
                $list = $_l;
            }
            return true;
        }

        public function checkYesNo( &$value, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $value ) )
            {
                return true;
            }
            if( in_array( strtolower( $value ), array( 'yes', 'no' ) ) )
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        public function checkInteger( &$value, $data = array() )
        {
            if( array_key_exists( 'empty', $data ) && $data['empty'] && empty( $value ) )
            {
                return true;
            }
            if( is_numeric( $value ) )
            {
                if( array_key_exists( 'min', $data ) || array_key_exists( 'max', $data ) )
                {
                    $min = array_key_exists( 'min', $data ) ? $data['min'] : -PHP_INT_MAX;
                    $max = array_key_exists( 'max', $data ) ? $data['max'] : PHP_INT_MAX;
                    if( (int)$value < $min || (int)$value > $max )
                    {
                        return false;
                    }
                }
                return true;
            }
            return false;
        }

    }
