<?php
    require_once (dirname( __FILE__ ) . "/spyc.php");

    class KYamlException extends Exception
    {
    }

    interface KYamlValidator
    {
        /**
         * Return validator name
         *
         * @return string name
         */
        public function getName();
        /**
         * Check if $data is valid and return true if it is, anything else
         * otherwise.
         * Throw exception in case of invalid data
         *
         * @param KYaml $yaml YAML object being validated
         * @param array $rules Current rule set (as in KYaml::validate)
         * @param mixed $data Current data (as in KYaml::validate)
         * @param string $path Parent
         */
        public function validate( KYaml $yaml, &$rule, &$data, &$path );
    }

    class KYamlIntegerRangeValidator implements KYamlValidator
    {
        protected $min;
        protected $max;
        protected $myname;

        /**
         * @param int $min Minimum acceptable value
         * @param int $max Maximum acceptable value
         */
        public function __construct( $min = PHP_INT_MIN, $max = PHP_INT_MAX, $myname = null )
        {
            $this->min = $min;
            $this->max = $max;
            $this->myname = is_null( $myname ) ? 'IntegerRange' : $myname;
        }
        /**
         * Return validator name
         *
         * @return string name
         */
        public function getName()
        {
            return $this->myname;
        }
        /**
         * Check if $data is valid and return true if it is, anything else
         * otherwise.
         * Throw exception in case of invalid data
         *
         * @param KYaml $yaml YAML object being validated
         * @param array $rules Current rule set (as in KYaml::validate)
         * @param mixed $data Current data (as in KYaml::validate)
         * @param string $path Parent
         */
        public function validate( KYaml $yaml, &$rule, &$data, &$path )
        {
            if( $data < $this->min || $data > $this->max )
            {
                $app = KSystem::getInstance( );
                throw KYamlIntegerRangeException( "{$app->translate( 'kyaml-integer-out-of-range' )}, {$data} [{$this->min}..{$this->max}]" );
            }
        }
    }

    class KYaml extends Spyc
    {
        //TODO: *should* be protected!!!
        public $yaml;
        /**
         * string File name and full path
         */
        protected $file;
        /**
         * string Backup file name and full path
         */
        protected $backupfile = null;

        //@formatter:off
        /**
         * array Validation rules
         *
         * Example YAML:
         *
         * ---
         * MyList:
         *   -
         *     Name: MyName
         *     Links:
         *       "0":
         *         Foo: Bar
         *   -
         *     Name: MyName2
         *     Links:
         *       "0":
         *         Foo: Bar2
         *       "1":
         *         Foo: Bar3
         *  MyRootKey2:
         *    Foo: ""
         *    Bar:
         *      -
         *        Test: true
         *        Debug: false
         *
         * Rules will be a set of neasted associative arrays defining some
         * properties of the elements.
         *
         * Valid properties are:
         *
         * type          - Data type (required)
         * children      - Valid children tags
         *                 (required, if type is list*, ignored otherwise)
         * default       - Default value
         * flags         - Flags
         *
         * Valid data types are:
         *
         * integer       - Integers
         * float         - Floats
         * string        - Strings
         * boolean       - Booleans
         * list          - Lists don't preserve index numbers
         * listhash      - Hash named list (keys are known, if the keys are
         *                 unknown or shouldn't be validated, use listindex)
         *
         * Valid flags are:
         *
         * FLAG_EMPTY    - The value can be empty
         * FLAG_NOTREQUIRED
         *               - This key is not required
         *
         * Default property values:
         *
         * default       - (undefined -- no default)
         * flags         - 0 (no flags set, value required and not empty)
         */
        protected $rules = array();
        //@formatter:on

        // validation constants
        const TYPE_INTEGER = 0;
        const TYPE_FLOAT = 1;
        const TYPE_STRING = 2;
        const TYPE_BOOLEAN = 4;
        const TYPE_LIST = 5;
        const TYPE_LISTHASH = 6;
        // flags
        const FLAG_NONE = 0;
        const FLAG_EMPTY = 1;
        const FLAG_NOTREQUIRED = 2;
        //POG: list normalization flag
        const FLAG_NORMALIZE = 4;

        // from Spyc::_dumpNode
        static protected $booleans = array( 'true', 'TRUE', 'false', 'FALSE', 'y', 'Y', 'n', 'N', 'null', 'NULL' );

        /**
         * string External validation program
         *
         * If set, then the KYaml::save method will use a two phase save:
         *
         * 1. Rename the configuration file to $this->file . ".kwebportal"
         * 2.1. Save the configuration
         * 2.2. Run validationProgram
         * 2.3. Check validation program exit code
         * 2.3.1. If zero, unlink $this->file . ".kwebportal"
         * 2.3.2. Else, rename $this->file . ".kwebportal" to $this->file
         *        and throw exception with the output buffer
         *
         * WARNING: stderr will always be redirected to stdout, so never supply
         *          another redirection pattern to stderr
         */
        protected $validationProgram = null;

        protected $validators = array();

        public function __construct( $file )
        {
            $this->file = $file;
            $this->yaml = $this->loadFile( $file );
        }

        /**
         * Recursively create default array values from rules
         *
         * @param array $rules
         * @return array
         */
        protected function _genarateDefault( &$rules, $path )
        {
            if( isset( $rules['children'] ) && ! isset( $rules['default'] ) )
            {
                $default = array( );

                foreach( $rules['children'] as $name => &$child )
                {
                    if(
                        ! isset( $child['default'] )
                        && isset( $child['flags'] )
                        && ( $child['flags'] & KYaml::FLAG_NOTREQUIRED ) == KYaml::FLAG_NOTREQUIRED )
                    {
                        continue;
                    }
                    $default[$name] = $this->_genarateDefault( $child, "{$path}.{$name}" );
                }

                return $default;
            }
            elseif( isset( $rules['default'] ) )
            {
                return $rules['default'];
            }

            trigger_error( get_class( $this ) . '::' . __FUNCTION__ . ": {$path} has no default value! Returning null.", E_USER_NOTICE );
            return null;
        }

        /**
         * Create default array values from rule path
         *
         * @param string $path
         * @return array
         */
        public function defaultFromPath( $path )
        {
            if( empty( $this->rules ) )
            {
                trigger_error( get_class( $this ) . '::' . __FUNCTION__ . ": No rules defined to create default array values from path {$path}", E_USER_NOTICE );
                return array( );
            }

            // find root node
            $parts = explode( '.', $path );
            if( ! isset( $this->rules[$parts[0]] ) )
            {
                trigger_error( get_class( $this ) . '::' . __FUNCTION__ . ": Root element {$part} from path {$path} not found", E_USER_NOTICE );
                return array( );
            }

            $myrules = & $this->rules[$parts[0]];
            array_shift( $parts );

            foreach( $parts as $part )
            {
                $found = false;

                if( ! isset( $myrules['children'] ) )
                {
                    trigger_error( get_class( $this ) . '::' . __FUNCTION__ . ": Element {$part} from path {$path} has no children elements", E_USER_NOTICE );
                    return array( );
                }

                foreach( $myrules['children'] as $name => &$child )
                {
                    if( $name === $part )
                    {
                        $myrules = &$child;
                        $found = true;
                        break;
                    }
                }

                if( ! $found )
                {
                    trigger_error( get_class( $this ) . '::' . __FUNCTION__ . ": Element {$part} not found in path {$path}", E_USER_NOTICE );
                    return array( );
                }
            }

            return $this->_genarateDefault( $myrules, $path );
        }


        /**
         * Validate the YAML against the internal validation rules
         */
        public function validate( $rules, &$data, $checkKeys = false, $parent = '' )
        {
            // only arrays can be validated
            if( !is_array( $data ) )
            {
                trigger_error( get_class( $this ) . ": Invalid data: '{$parent}[unknown]'", E_USER_WARNING );
                throw new KYamlException( 'kyaml-error-runtime-invalid-data' );
            }
            elseif( !is_array( $rules ) )
            {
                trigger_error( get_class( $this ) . ": Invalid rules: '{$parent}[unknown]'", E_USER_WARNING );
                throw new KYamlException( 'kyaml-error-runtime-invalid-rules' );
            }

            // keys seen, will be used to check missing keys in pass #2
            $seen = array( );
            // pass #1
            foreach( $data as $key => &$item )
            {
                // required items are checked in pass #2
                if( is_null( $item ) )
                {
                    continue;
                }
                if( array_key_exists( $key, $rules ) )
                {
                    $rule = &$rules[$key];
                    $seen[] = $key;
                }
                // undefined key name, should be used *inside* lists
                elseif( $checkKeys === false && array_key_exists( '?', $rules ) )
                {
                    $rule = &$rules['?'];
                }
                else
                {
                    // keys without validation rules are ignored
                    continue;
                }

                // check empty
                if( $item === '' || (is_array( $item ) && count( $item ) === 0) )
                {
                    if( $this->isValidationFlagSet( $rule, KYaml::FLAG_EMPTY ) )
                    {
                        continue;
                    }
                    elseif( array_key_exists( 'default', $rule ) )
                    {
                        $item = $rule['default'];
                        continue;
                    }
                    else
                    {
                        trigger_error( get_class( $this ) . ": Key '{$parent}{$key}' cannot be empty", E_USER_WARNING );
                        throw new KYamlException( 'kyaml-error-empty-key' );
                    }
                }

                switch( $rule['type'] )
                {
                    case KYaml::TYPE_INTEGER:
                        if( is_string( $item ) && preg_match( '/^[\-]?[0-9]+$/', trim( $item ) ) )
                        {
                            $item = (int)$item;
                        }
                        elseif( !is_int( $item ) )
                        {
                            trigger_error( get_class( $this ) . ": Key '{$parent}{$key}' has an invalid value '{$item}', integer expected", E_USER_WARNING );
                            throw new KYamlException( 'kyaml-error-integer-expected' );
                        }
                        break;
                    case KYaml::TYPE_FLOAT:
                        if( is_string( $item ) && preg_match( '/^[\-]?[0-9][0-9\.]+[0-9]$/', trim( $item ) ) )
                        {
                            $item = floatval( $item );
                        }
                        elseif( !is_float( $item ) )
                        {
                            trigger_error( get_class( $this ) . ": Key '{$parent}{$key}' has an invalid value '{$item}', float expected", E_USER_WARNING );
                            throw new KYamlException( 'kyaml-error-float-expected' );
                        }
                        break;
                    case KYaml::TYPE_STRING:
                        if( is_float( $item ) || is_integer( $item ) )
                        {
                            $item = (string)$item;
                        }
                        elseif( is_bool( $item ) )
                        {
                            $item = $item ? 'true' : 'false';
                        }
                        elseif( !is_string( $item ) )
                        {
                            trigger_error( get_class( $this ) . ": Key '{$parent}{$key}' has an invalid value '{$item}', string expected", E_USER_WARNING );
                            throw new KYamlException( 'kyaml-error-string-expected' );
                        }
                        break;
                    case KYaml::TYPE_BOOLEAN:
                        if( !((is_string( $item ) && in_array( $item, self::$booleans )) || is_bool( $item )) )
                        {
                            trigger_error( get_class( $this ) . ": Key '{$parent}{$key}' has an invalid value '{$item}', boolean expected", E_USER_WARNING );
                            throw new KYamlException( 'kyaml-error-boolean-expected' );
                        }
                        // cast to boolean
                        elseif( is_string( $item ) )
                        {
                            switch( $item )
                            {
                                case 'true':
                                case 'TRUE':
                                case 'y':
                                case 'Y':
                                    $item = true;
                                    break;
                                case 'false':
                                case 'FALSE':
                                case 'n':
                                case 'N':
                                case 'null':
                                case 'NULL':
                                    $item = false;
                                    break;
                            }
                        }
                        break;
                    case KYaml::TYPE_LISTHASH:
                        // recursion
                        $this->validate( $rule['children'], $item, true, "{$parent}{$key} => " );
                        break;
                    case KYaml::TYPE_LIST:
                        if( is_array( $item ) )
                        {
                            if( $this->isValidationFlagSet( $rule, KYaml::FLAG_NORMALIZE ) )
                            {
                                //WARNING: $item is passed by reference here
                                if( !$this->normalize( $item ) )
                                {
                                    trigger_error( get_class( $this ) . ": Normalization error, {$parent}{$key} is not a numeric list?", E_USER_WARNING );
                                }
                            }
                            $this->validate( $rule['children'], $item, false, "{$parent}{$key} => " );
                        }
                        else
                        {
                            trigger_error( get_class( $this ) . ": '{$parent}{$key}' is not an array, skipping recursion", E_USER_WARNING );
                        }
                        break;
                    default:
                        trigger_error( get_class( $this ) . ": Key '{$parent}{$key}' has an invalid configuration, check this->rules array", E_USER_WARNING );
                        throw new KYamlException( 'kyaml-error-invalid-type' );
                        break;
                }

                // extended check
                if( isset( $rule['validator'] ) )
                {
                    $this->validateEx( $rule, $item, "{$parent}{$key}" );
                }
            }
            //pass #2: check required keys
            if( $checkKeys )
            {
                $missing = array_diff( array_keys( $rules ), $seen );
                // if there are missing keys, check if they are required
                if( !empty( $missing ) )
                {
                    foreach( $missing as $key )
                    {
                        $required = !$this->isValidationFlagSet( $rules[$key], KYaml::FLAG_NOTREQUIRED );
                        if( $required && !array_key_exists( 'default', $rules[$key] ) )
                        {
                            trigger_error( get_class( $this ) . ": Missing required key '{$parent}{$key}'", E_USER_WARNING );
                            throw new KYamlException( 'kyaml-error-missing-key' );
                        }
                        elseif( array_key_exists( 'default', $rules[$key] ) )
                        {
                            $data[$key] = $rules[$key]['default'];
                        }
                    }
                }
            }
        }

        /**
         * Call validator(s)
         */
        protected function validateEx( &$rule, &$data, $path )
        {
            if( is_array( $rule['validator'] ) )
            {
                foreach( $validators as $validator )
                {
                    $this->callValidator( $validator, $rule, $data, $path );
                }
            }
            else
            {
                $this->callValidator( $rule['validator'], $rule, $data, $path );
            }
        }

        /**
         * Call validator by name
         */
        protected function callValidator( $name, &$rule, &$data, $path )
        {
            if( isset( $this->validators[$name] ) )
            {
                try
                {
                    $this->validators[$name]->validate( $this, $rules, $data, $path );
                }
                catch( Exception $e )
                {
                    trigger_error( get_class( $this ) . ": '{$data}' for key '{$path}' rejected by validator '{$name}': {$e->getMessage()}", E_USER_WARNING );
                    throw $e;
                }
            }
            else
            {
                trigger_error( get_class( $this ) . ": Missing validator '{$name}', ignoring it during validation of '{$path}'", E_USER_WARNING );
            }
        }

        /**
         * Add a custom validator, suitable for "validator" rule key
         *
         * @param KYamlValidator $validator Validator instance
         */
        protected function addValidator( KYamlValidator $validator )
        {
            $this->validators[$validator->getName()] = $validator;
        }

        /**
         * Check if flag is set in rule flags
         *
         * @param array $rule Rule array
         * @param int $flag Flag to be checked
         * @return boolean
         */
        protected function isValidationFlagSet( $rule, $flag )
        {
            return array_key_exists( 'flags', $rule ) && (($rule['flags'] & $flag) === $flag);
        }

        /**
         * Numeric array normalization
         *
         * @param array $data
         */
        protected function normalize( &$data )
        {
            if( is_array( $data ) )
            {
                foreach( array_keys( $data ) as $key )
                {
                    if( !is_int( $key ) )
                    {
                        // error: has keys that aren't int!!!
                        return false;
                    }
                }
                // remove intervals
                $node = array_values( $data );
                // apply normalized data
                $data = $node;
                return true;
            }
            return false;
        }

        /**
         * Save YAML
         */
        public function save( )
        {
            $this->setting_dump_force_quotes = true;
            $this->validate( $this->rules, $this->yaml, true );

            $this->backup( );
            $this->writeYaml( );
            $this->commit( );
        }

        /**
         * Backup YAML
         *
         * @throws KYamlException
         */
        protected function backup( )
        {
            if( !is_null( $this->backupfile ) )
            {
                throw new KYamlException( 'kyaml-error-already-backuped' );
            }
            if( !is_null( $this->validationProgram ) )
            {
                $this->backupfile = "{$this->file}.kwebportal";
                if( rename( $this->file, $this->backupfile ) === false )
                {
                    throw new KYamlException( 'kyaml-error-write-failed' );
                }
            }
        }

        /**
         * Generate new YAML file
         *
         * @throws KYamlException
         */
        protected function writeYaml( )
        {
            if( file_put_contents( $this->file, $this->getHeader( ) . $this->dump( $this->yaml, false, 0 ), LOCK_EX ) === false )
            {
                $this->rollback( );
                throw new KYamlException( 'kyaml-error-write-failed' );
            }
        }

        /**
         * Validate current YAML file using validationProgram and commit new
         * configuration by removing backup file or rollback the backup file if
         * validation fail
         *
         * @throws KYamlException
         */
        protected function commit( )
        {
            if( !is_null( $this->backupfile ) )
            {
                $output = array( );
                $errcode = -1;
                // external validation
                exec( "{$this->validationProgram} 2>&1", $output, $errcode );
                if( $errcode == 0 )
                {
                    // EX_OK, commit the new file by unlinking the backup
                    unlink( $this->backupfile );
                    $this->backupfile = null;
                }
                else
                {
                    // Validation error, roolback and throw exception
                    $this->rollback( );
                    throw new KYamlException( implode( ' ', $output ) );
                }
            }
        }

        /**
         * Restore backuped YAML
         *
         * @throws KYamlException
         */
        protected function rollback( )
        {
            if( !is_null( $this->backupfile ) )
            {
                $backupfile = $this->backupfile;
                $this->backupfile = null;
                if( rename( $backupfile, $this->file ) === false )
                {
                    throw  KYamlException( 'kyaml-error-write-failed' );
                }
            }
        }

        /**
         * YAML header
         *
         * @return string
         */
        protected function getHeader( )
        {
            $text = "# Khomp configuration file\n";
            $text .= "# Please note that indentation denotes scope but TABs are not permitted\n";
            $text .= "# All fields are case sensitive\n\n";

            return $text;
        }

        ///////////////////////////////////////////////////////////////////////
        // Common stuff
        ///////////////////////////////////////////////////////////////////////

        /**
         * Check if the given name is valid
         *
         * @param string $name Name to be tested
         * @return boolean
         */
        public function isValidName( $name )
        {
            return preg_match( '/[a-z]/i', $name ) === 1;
        }

        ///////////////////////////////////////////////////////////////////////
        // OLD stuff
        ///////////////////////////////////////////////////////////////////////

        /**
         * Save YAML old style
         *
         * @deprecated use KYaml::save instead
         */
        public function saveYaml( $FormatIndexes = false )
        {
            $fp = fopen( $this->file, "w+b" );
            $text = $this->getHeader( );
            $text .= $this->dump( $this->yaml );
            //$text .= $this->fileFormatYamlData( );

            fwrite( $fp, $text );
            fclose( $fp );
            $fileData = file( $this->file );
            $dataSize = sizeof( $fileData );
            for( $i = 0; $i < $dataSize; $i++ )
                $fileData[$i] = rtrim( $fileData[$i] );
            $text = '';
            $text = implode( PHP_EOL, $fileData );
            $fp = fopen( $this->file, "w+b" );
            fwrite( $fp, $text );
            fclose( $fp );

            $this->keyParser( );
            if( $FormatIndexes )
                $this->extraParser( );
        }

        function keyParser( )
        {
            $yamlContent = file( $this->file );
            $fileSize = sizeof( $yamlContent );
            for( $x = 0; $x < $fileSize; $x++ )
            {
                if( @ereg( ":", $yamlContent[$x] ) )
                {
                    $tmp = explode( ":", $yamlContent[$x] );
                    if( trim( $tmp[1] ) == "" )
                    {
                        $tmp[0] = $this->rebuildLine( $tmp[0] );
                        $yamlContent[$x] = $tmp[0] . ":" . $tmp[1];
                    }
                }
            }
            $fp = fopen( $this->file, "w+b" );
            fwrite( $fp, implode( "", $yamlContent ) );
            fclose( $fp );
        }

        function rebuildLine( $piece )
        {
            $pieceLength = strlen( $piece );
            $first = true;
            $newStr = '';
            if( $piece[$pieceLength - 1] == "\"" )
                return $piece;
            for( $i = 0; $i < $pieceLength; $i++ )
            {
                if( $piece[$i] == " " )
                    $newStr .= $piece[$i];
                else
                {
                    if( $first )
                    {
                        $first = false;
                        $newStr .= "\"" . $piece[$i];
                    }
                    else
                        $newStr .= $piece[$i];
                }
            }
            return $newStr . "\"";
        }

        function extraParser( )
        {
            $yamlContent = file( $this->file );
            $fileSize = sizeof( $yamlContent );
            $structureStart = false;
            for( $x = 0; $x < $fileSize; $x++ )
            {
                if( $structureStart && strpos( $yamlContent[$x], ' -' ) !== false )
                {
                    $yamlContent[$x] = str_replace( '-', '"0":', $yamlContent[$x] );
                    $structureStart = false;
                }
                else
                {
                    $structureStart = strpos( $yamlContent[$x], ' -' ) !== false;
                }
            }
            $fp = fopen( $this->file, "w+b" );
            fwrite( $fp, implode( "", $yamlContent ) );
            fclose( $fp );
        }

        function getYaml( )
        {
            return $this->yaml;
        }

        function setYaml( $arr )
        {
            $this->yaml = $arr;
        }

        function fileFormatYamlData( )
        {
            return Spyc::YAMLDump( $this->yaml );
        }

        function fileFormatFromYamlData( $yamlData )
        {
            return Spyc::YAMLDump( $yamlData );
        }

        function getWay( $yaml = "", $varName, $way = "" )
        {
            if( $yaml == "" )
                $yaml = $this->yaml;
            $x = 0;
            $keys = array_keys( $yaml );

            foreach( $yaml as $obj )
            {
                if( is_array( $obj ) )
                {
                    $tempWay = $way;
                    if( $way == "" )
                        $way = $keys[$x];
                    else
                        $way .= "|" . $keys[$x];
                    if( array_key_exists( $varName, $obj ) )
                        return $way;
                    else
                    {
                        if( $this->getWay( $obj, $varName, $way ) )
                            return $this->getWay( $obj, $varName, $way );
                    }
                    $way = $tempWay;
                }
                $x++;
            }
            return false;
        }

        function createVar( $varName, $way, $value )
        {
            if( $way )
                $this->createData( $this->yaml, $way . "|" . $varName, $value );
        }

        function createData( &$data, $keySequence, $value )
        {
            $keys = explode( "|", $keySequence );
            $ckey = $keys[0];
            array_shift( $keys );
            if( sizeof( $keys ) == 0 )
            {
                if( is_array( $data ) )
                    $data[$ckey] = $value;
                else
                    $data = array( $ckey => $value );
            }
            else
            {
                $keys = implode( "|", $keys );
                $this->createData( $data[$ckey], $keys, $value );
            }
        }

        function createSubVar( $varName, $subVarName, $way, $value )
        {
            if( $way )
                $this->createData( $this->yaml, $way . "|" . $varName . "|" . $subVarName, $value );
        }

        function createAddValue( $varName, $way, $value )
        {
            if( $way )
                $this->createData( $this->yaml, $way . "|" . $varName, $value );
        }

        function deleteVar( $varName, $way = '' )
        {
            if( $way == '' )
                $way = $this->getWay( $this->yaml, $varName );
            if( $way )
                $this->deleteData( $this->yaml, $way . "|" . $varName );
        }

        function deleteSubVar( $varName, $subVarName, $way = '' )
        {
            if( $way == '' )
                $way = $this->getWay( $this->yaml, $varName );

            if( $way )
                $this->deleteData( $this->yaml, $way . "|" . $varName . "|" . $subVarName );
        }

        function sortKeys( $piece )
        {
            if( $piece != "" )
                $yamlSize = sizeof( $this->yaml[$piece] );
            else
                $yamlSize = sizeof( $this->yaml );
            $newYaml = array( );

            foreach( $this->yaml[$piece] as $slice )
                $newYaml[] = $slice;

            $this->yaml[$piece] = $newYaml;
        }

        function getValue( $varName, $way = "", $default = 'yamlError' )
        {
            if( $way == "" )
                $way = $this->getWay( $this->yaml, $varName );

            if( $way )
                return $this->getData( $this->yaml, $way . "|" . $varName, $default );

            return $default;
        }

        function changeVarValue( $varName, $value, $way = "" )
        {
            if( $way == "" )
                $way = $this->getWay( $this->yaml, $varName );

            if( $way )
                $this->changeData( $this->yaml, $way . "|" . $varName, $value );
            return;
        }

        function changeData( &$data, $keySequence, $value )
        {
            $keys = explode( "|", $keySequence );
            $ckey = $keys[0];
            array_shift( $keys );
            if( sizeof( $keys ) == 0 )
                $data[$ckey] = $value;
            else
            {
                $keys = implode( "|", $keys );
                $this->changeData( $data[$ckey], $keys, $value );
            }
        }

        function deleteData( &$data, $keySequence )
        {
            $keys = explode( "|", $keySequence );
            $ckey = $keys[0];
            array_shift( $keys );
            if( sizeof( $keys ) == 0 )
                unset( $data[$ckey] );
            else
            {
                $keys = implode( "|", $keys );
                $this->deleteData( $data[$ckey], $keys );
            }
        }

        function getData( &$data, $keySequence, $returnError = '' )
        {
            $keys = explode( "|", $keySequence );
            $ckey = $keys[0];
            array_shift( $keys );
            if( sizeof( $keys ) == 0 )
            {
                return isset( $data[$ckey] )
                    ? $data[$ckey]
                    : $returnError;
            }
            else
            {
                $keys = implode( "|", $keys );
                return isset( $data[$ckey] )
                    ? $this->getData( $data[$ckey], $keys, $returnError )
                    : $returnError;
            }
        }

    }
