(function (Khomp) {

    /// Low layer compability
    /// Página 81 da ITU Q931.

    var ISDN = Khomp.ISDN;
    var InformationElements = ISDN.InformationElements;
    var attributes = InformationElements.attributes;

    InformationElements.elementsDecoders[0x7c] = function (values) {

        /// Objeto com a informação decodificada que será retornado.
        var objValue = {};

        /// Octet 3

        var octet3 = values.shift();

        /// Coding standard: bits 7,6 octet 3
        var codingStandard = (octet3 & 0x60) >> 5;

        /// Information transfer capability: bits 5,4,3,2,1 octet 3
        var informationTransferCapability = (octet3 & 0x1f);

        objValue[InformationElements.attributeNames[0x06]] = attributes[0x06][codingStandard] || codingStandard;
        objValue['Information transfer capability'] = attributes[0x10][informationTransferCapability] || informationTransferCapability;

        /// Octet 3a

        /// Negotiation indicator: bit 7 octet 3a
        /// TABLE 4-16/Q.931 NOTE 4 – When octet 3a is omitted, "out-band negotiation not possible" shall be assumed.
        objValue['Negotiation indicator'] = attributes[0x1f][0x00];

        if (!InformationElements.isLastOctet(octet3)) {

            var octet3a = values.shift();
            var negotiationIndicator = (octet3a & 0x40) >> 6;

            objValue['Negotiation indicator'] = attributes[0x1f][negotiationIndicator] || negotiationIndicator;

        }

        /// Octet 4

        var octet4 = values.shift();

        /// Transfer mode: bits 7,6 octet 4
        var transferMode = (octet4 & 0x60) >> 5;

        /// Information transfer rate: bits 5,4,3,2,1 octet 4
        /// TABLE 4-16/Q.931 NOTE 5 – When the information transfer rate 2×64 kbit/s is used, the coding of octets 3 and 4 refer to both 64 kbit/s channels.
        /// TABLE 4-16/Q.931 NOTE 6 – Additional attributes are defined in Table 4-17
        var informationTransferRate = (octet4 & 0x1f);

        objValue['Transfer mode'] = attributes[0x34][transferMode] || transferMode;
        objValue['Information transfer rate'] = attributes[0x11][transferMode][informationTransferRate] || informationTransferRate;

        /// Octet 4.1*
        /// NOTE 1 – This octet is required if octet 4 indicates multirate (64 kbit/s base rate). Otherwise, it shall not be present.

        if (informationTransferRate === 0x18) { // multirate (64 kbit/s base rate)

            var octet41 = values.shift();

            /// Rate multiplier
            objValue['Rate multiplier'] = octet41;
        }

        /// Octet 5*

        if (values.length === 0) {

            /// low layer compability - Caiu nesta condição porque normalmente deve ter mais dados (CONFIRMAR ISSO!!)

            return objValue;
        }

        var octet5 = values.shift();
        var layer1Ident = (octet5 & 0x60) >> 5;

        /// User information layer 1 protocol: bits 5,4,3,2,1 octet 5
        var userInfoLayer1Protocol = (octet5 & 0x1f);

        objValue['Layer 1 identifier'] = layer1Ident;
        objValue['User information layer 1 protocol'] = attributes[0x38][userInfoLayer1Protocol] || userInfoLayer1Protocol;

        /// Octet 5a*
        /// NOTE 2 – This octet may be present
        /* if octet 3 indicates unrestricted digital information
        and octet 5 indicates either of the
        ITU - T standardized rate adaptions V.110,
        I.460
        and X.30 or V.120[9].
        It may also be present
        if (octet 3 indicates 3.1 kHz audio)
        and(octet 5 indicates G.711).
         */
        var conditionNote2 = (
            (informationTransferCapability === 0x11 && (userInfoLayer1Protocol === 0x01 || userInfoLayer1Protocol === 0x08)) ||
            (informationTransferCapability === 0x11 && (userInfoLayer1Protocol === 0x02 || userInfoLayer1Protocol === 0x03))) ? true : false;

        var extensionIndicatorOct5 = (((octet5 & 0x80) >> 7) === 0) ? true : false;

        if (!extensionIndicatorOct5) {
            return objValue;
        }

        var octet5a = values.shift();

        /// Synchronous/Asynchronous: bit 7 octet 5a
        var syncData = (octet5a & 0x40) >> 6;
        objValue['Synchronous/Asynchronous'] = attributes[0x33][syncData] || syncData;

        /// Negotiation: bit 6 octet 5a
        var negotiation = (octet5a & 0x20) >> 5;
        objValue.Negotiation = attributes[0x1e][negotiation] || negotiation;

        /// User rate: bits 5,4,3,2,1 octet 5a
        var userRate = (octet5a & 0x1f);
        objValue['User rate'] = attributes[0x3b][userRate] || userRate;

        /// Octet 5b*
        /// NOTE 3 – This structure of octet 5b only applies if octet 5 indicates ITU-T standardized rate adaption (see Recommendations V.110 [7], I.460 [15] and X.30 [8])
        if (values.length && userInfoLayer1Protocol === 0x01) {

            var octet5b = values.shift();

            /// Intermediate rate bits 7,6 octet 5b
            var intermediateRate = (octet5b & 0x60) >> 5;
            objValue['Intermediate rate'] = attributes[0x14][intermediateRate] || intermediateRate;

            /// NIC on Tx: bit 5 octet 5b
            var nicOnTx = (octet5b & 0x10) >> 4;
            objValue['NIC on Tx'] = attributes[0x22][nicOnTx] || nicOnTx;

            /// NIC on Rx: bit 4 octet 5b
            var nicOnRx = (octet5b & 0x08) >> 3;
            objValue['NIC on Rx'] = attributes[0x21][nicOnRx] || nicOnRx;

            /// Flow control on Tx: bit 3 octet 5b
            var flowControlOnTx = (octet5b & 0x04) >> 2;
            objValue['Flow control on Tx'] = attributes[0x0c][flowControlOnTx] || flowControlOnTx;

            /// Flow control on Rx: bit 2 octet 5b
            var flowControlOnRx = (octet5b & 0x02) >> 1;
            objValue['Flow control on Rx'] = attributes[0x0b][flowControlOnRx] || flowControlOnRx;

            /// NOTE 4 – This structure of octet 5b only applies if octet 5 indicates ITU-T standardized rate adaption (see Recommendation V.120 [9]).
            if (!InformationElements.isLastOctet(octet5b) && userInfoLayer1Protocol === 0x08) {

                var octet5ba = values.shift();

                /// Rate adaption header: bit 7 octet 5b*
                var rateAdaptionHeader = (octet5ba & 0x40) >> 6;
                objValue['Rate adaption header'] = attributes[0x30][rateAdaptionHeader] || rateAdaptionHeader;

                /// Multiframe: bit 6 octet 5b*
                var multiframe = (octet5ba & 0x20) >> 5;
                objValue.Multiframe = attributes[0x1d][multiframe] || multiframe;

                /// Mode of operation: bit 5 octet 5b*
                var modeOfOperation = (octet5ba & 0x10) >> 4;
                objValue['Mode of operation'] = attributes[0x19][modeOfOperation] || modeOfOperation;

                /// Logical link identifier negotiation: bit 4 octet 5b*
                var logicalLinkIdentifierNegotiation = (octet5ba & 0x08) >> 3;
                objValue['Logical link identifier negotiation'] = attributes[0x17][logicalLinkIdentifierNegotiation] || logicalLinkIdentifierNegotiation;

                /// Assignor/assignee: bit 3 octet 5b*
                var assignorAssignee = (octet5ba & 0x04) >> 2;
                objValue[InformationElements.attributeNames[0x01]] = attributes[0x01][assignorAssignee] || assignorAssignee;

                /// In-band/out-band negotiation: bit 2 octet 5b*
                var inBandOutBandNegotiation = (octet5ba & 0x02) >> 1;
                objValue['In-band/out-band negotiation'] = attributes[0x0e][inBandOutBandNegotiation] || inBandOutBandNegotiation;

                /// Octet 5c*
                /// NOTE 2
                if (values.length && !InformationElements.isLastOctet(octet5ba) && conditionNote2) {

                    var octet5c = values.shift();

                    /// Number of stop bits: bits 7,6 octet 5c*
                    var numberOfStopBits = (octet5c & 0x60) >> 5;
                    objValue['Number of stop bits'] = attributes[0x26][numberOfStopBits] || numberOfStopBits;

                    /// Number of data bits excluding parity Bit if present: bits 5,4 octet 5c*
                    var numberOfDataBitsExcludingParityBitIfPresent = (octet5c & 0x18) >> 3;
                    objValue['Number of data bits excluding parity Bit if present'] = attributes[0x25][numberOfDataBitsExcludingParityBitIfPresent] || numberOfDataBitsExcludingParityBitIfPresent;

                    /// Parity information: bits 3,2,1 octet 5c*
                    var parityInformation = (octet5c & 0x07);
                    objValue['Parity information'] = attributes[0x29][parityInformation] || parityInformation;

                    /// Octet 5d*
                    /// NOTE 2
                    if (values.length && !InformationElements.isLastOctet(octet5c) && conditionNote2) {

                        var octet5d = values.shift();

                        /// Mode duplex: bit 7 octet 5d*
                        var modeDuplex = (octet5d & 0x40) >> 6;
                        objValue['Mode duplex'] = attributes[0x18][modeDuplex] || modeDuplex;

                        /// Modem type: bits 6,5,4,3,2,1 octet 5d*
                        var modemType = (octet5d & 0x3f);
                        objValue['Modem type'] = attributes[0x1c][modemType] || modemType;

                    }
                }
            }

            /// Octet 6*
            if (values.length) {

                var octet6 = values.shift();

                var layer2Ident = (octet6 & 0x60) >> 5;
                objValue['Layer 2 identifier'] = layer2Ident;

                /// User information layer 2 protocol:  bits 5,4,3,2,1 octet 6
                /// NOTE 20 – This Recommendation is compatible with ISO/IEC 7776 DTE-DCE operation.
                /// NOTE 21 – When this coding is included, octet 6a will include user coding for the user specified Layer 2 protocol.
                /// NOTE 22 – This Standard is compatible with Recommendation X.75 modified by the application rules defined in Recommendation T.90.
                /// NOTE 23 – When this coding is included, octets 6a and 6b with ITU-T encoding may be included.
                var userInfoLayer2Protocol = (octet6 & 0x1f);
                objValue['User information layer 2 protocol'] = attributes[0x39][userInfoLayer2Protocol] || userInfoLayer2Protocol;

                /// NOTE 5 – This octet may be present only if octet 6 indicates certain acknowledged mode HDLC elements of procedure as indicated in Table 4-16.
                if (!InformationElements.isLastOctet(octet6)) {

                    /// NOTE 23 – When this coding is included, octets 6a and 6b with ITU-T encoding may be included.

                    if (userInfoLayer2Protocol === 0x02 ||
                        userInfoLayer2Protocol === 0x06 ||
                        userInfoLayer2Protocol === 0x07 ||
                        userInfoLayer2Protocol === 0x09 ||
                        userInfoLayer2Protocol === 0x0a ||
                        userInfoLayer2Protocol === 0x0b ||
                        userInfoLayer2Protocol === 0x0d ||
                        userInfoLayer2Protocol === 0x0e ||
                        userInfoLayer2Protocol === 0x11) {

                        /// Octet 6a*
                        var octet6a = values.shift();

                        /// Mode of operation: bits 7,6 octet 6a
                        var modeOfOperationLayer2 = (octet6a & 0x60) >> 5;

                        objValue['Mode of operation layer 2'] = attributes[0x1a][modeOfOperationLayer2] || modeOfOperationLayer2;

                        /// Q.933 use: bits 2,1 octet 6a
                        var q933 = (octet6a & 0x03);

                        objValue['Q.933 use'] = attributes[0x2f][q933] || q933;

                        if (!InformationElements.isLastOctet(octet6a)) {

                            /// Octet 6a2*
                            var octet6a2 = values.shift();
                            var userSpecifiedLayer2ProtocolInformation = (octet6a2 & 0x7f);
                            objValue['User specified layer 2 protocol information '] = userSpecifiedLayer2ProtocolInformation;

                        }
                        /// Octet 6b*
                        var octet6b = values.shift();
                        var windowSize = (octet6b & 0x7f);
                        objValue['Window size'] = windowSize + 'k';
                    }

                    /// Octet 7
                    var octet7 = values.shift();

                    var layer3Identifier = (octet7 & 0x60) >> 5;
                    objValue['Layer 3 identifier'] = layer3Identifier;

                    /// User information layer 3 protocol: bits 5,4,3,2,1 octet 7
                    var userInfoLayer3Protocol = (octet7 & 0x1f);
                    objValue['User information layer 3 protocol'] = attributes[0x3a][userInfoLayer3Protocol] || userInfoLayer3Protocol;

                    /// Octet 7a* - primeiro caso
                    var octet7a;

                    if (userInfoLayer3Protocol === 0x10) {
                        /// NOTE 8 – This octet may be present only if octet 7 indicates 'user specified' layer 3 protocol.
                        /// Table 4-16/Q.931 NOTE 24 – When this coding is included, octet 7a will include user coding for the user specified layer 3 protocol.

                        octet7a = values.shift();

                        objValue['Optional layer 3 protocol information'] = (octet7a);

                        /// Octet 7a* - segundo caso
                    } else if (
                        /// NOTE 7 – This octet may be present only if octet 7 indicates a layer 3 protocol based on ITU-T X.25 [5], ISO/IEC 8208 [41] or ITU-T Rec. X.223 [96] and ISO/IEC 8878 [81] as indicated in Table 4-16.
                        /// Table 4-16/Q.931 NOTE 25 – When this coding is included, octets 7a, 7b and 7c with ITU-T X.25, X.223 and ISO/IEC TR 9577 encoding may be included.

                        userInfoLayer3Protocol === 0x06 ||
                        userInfoLayer3Protocol === 0x07 ||
                        userInfoLayer3Protocol === 0x08) {

                        octet7a = values.shift();

                        /// Mode of operation layer 3: bits 7,6 octet 7a
                        var modeOfOperationLayer3 = (octet7a & 0x60) >> 5;

                        objValue['Mode of operation layer 3'] = attributes[0x1b][modeOfOperationLayer3] || modeOfOperationLayer3;

                        if (!InformationElements.isLastOctet(octet7a)) {

                            var octet7b = values.shift();

                            /// Default packet size: bits 4,3,2,1 octet 7b
                            var defaultPacketSize = (octet7b & 0x0f);

                            objValue['Default packet size'] = attributes[0x08][defaultPacketSize] || defaultPacketSize;

                            if (!InformationElements.isLastOctet(octet7b)) {

                                var octet7c = values.shift();
                                /// Packet window size: octet7c bits 7,6,5,4,3,2,1
                                objValue['Packet window size'] = (octet7c & 0x7f);
                            }
                        }

                    } else if (userInfoLayer3Protocol === 0x0b) {
                        /// Octet 7a* - terceiro caso
                        /// NOTE 9 – This may be included if octet 7 indicates ISO/IEC TR 9577.
                        /// NOTE 26 – When this coding is included, octets 7a and 7b with ITU-T X.25, X.223 [96] and ISO/IEC TR 9577 encoding may be included
                    }

                }
            }
        }

        if (values.length) {
            objValue['Não decodificado'] = values.join(' ');
        }
        return objValue;
    };

})(Khomp);
