<?php
    require_once KROOT . '/classes/KSystem.php';

    class KLoggerException extends Exception
    {
    }

    /**
     * Logger configuration
     *
     * Trace lavel is enable only by FullLog
     */
    class KLoggerConfig
    {
        static private $_this = null;
        protected $trace = false;

        /**
         * Read the configuration file
         */
        private function __construct( )
        {
            $app = KSystem::getInstance( );
            $cfgfile = fopen( "{$app->getConfigDir()}/klog.cfg", 'r' );
            if( $cfgfile )
            {
                while( ($line = fgets( $cfgfile )) !== false )
                {
                    if( stripos( $line, 'fulllog' ) !== false )
                    {
                        $parts = explode( '=', trim( $line ) );
                        if( count( $parts ) == 2 )
                        {
                            $this->trace = strncasecmp( 'true', trim( $parts[1] ), 4 ) === 0;
                            break;
                        }
                    }
                }
                fclose( $cfgfile );
            }
        }

        /**
         * Singleton implementation
         *
         * @return KLoggerConfig
         */
        static public function getInstance( )
        {
            if( is_null( self::$_this ) )
            {
                self::$_this = new self( );
            }
            return self::$_this;
        }

        /**
         * Check if trace level is enabled
         *
         * @return boolean
         */
        public function isTraceEnabled( )
        {
            return $this->trace;
        }

    }

    /**
     * Logging facility
     */
    class KLogger
    {
        protected $filename;
        protected $path;
        protected $source;

        /**
         * @param string $filename File name to be written to (only the name,
         *     no path or extension)
         * @param string $source Logger source
         */
        public function __construct( $filename, $source = 'system' )
        {
            $this->filename = $filename;
            $this->source = strtoupper( $source );
            if( PHP_OS == 'WINNT' )
            {
                throw new KLoggerException( "Windows support is not implemented yet!" );
                //TODO: $this->path = "c:\\khomp\\logs\\";
            }
            else
            {
                $this->path = '/var/log/khomp/';
            }
        }

        /**
         * Write message to file
         *
         * @param string $filename File name as in constructor
         * @param string $msg Message to be written
         */
        protected function __write( $filename, $msg )
        {
            $handler = fopen( "{$this->path}{$filename}.log", "a" );
            if( $handler )
            {
                fwrite( $handler, $msg );
                fclose( $handler );
            }
        }

        /**
         * Format and call write to file
         *
         * @param string $level Message level
         * @param array $args Message format (array position 0) and format
         *     arguments (positions 1..n)
         * @param boolean $messages If true, the message is also written to
         *     messages.log file
         */
        protected function write( $level, $args, $messages = false )
        {
            $fmt = $args[0];
            unset( $args[0] );
            $header = sprintf( '%s|%s.%s|%-8s', $level, strftime( '%d/%m/%Y %H:%M:%S' ), substr( microtime( ), 2, 3 ), $this->source );
            $msg = vsprintf( "{$header}| {$fmt}", $args );
            $this->__write( $this->filename, "{$msg}\n" );
            if( $messages )
            {
                $this->__write( 'messages', "{$msg} ({$this->filename})\n" );
            }
        }

        /**
         * Issue trace level messages
         *
         * @param string $fmt Format
         * @param mixed ... Format values (variadic)
         */
        public function trace( $fmt )
        {
            if( !KLoggerConfig::getInstance( )->isTraceEnabled( ) )
            {
                return;
            }
            $args = func_get_args( );
            $this->write( 'T', $args );
        }

        /**
         * Issue notice level messages
         *
         * @param string $fmt Format
         * @param mixed ... Format values (variadic)
         */
        public function notice( $fmt )
        {
            $args = func_get_args( );
            $this->write( 'N', $args );
        }

        /**
         * Issue info level messages (also goes to messages.log)
         *
         * @param string $fmt Format
         * @param mixed ... Format values (variadic)
         */
        public function info( $fmt )
        {
            $args = func_get_args( );
            $this->write( 'I', $args, true );
        }

        /**
         * Issue warning level messages (also goes to messages.log)
         *
         * @param string $fmt Format
         * @param mixed ... Format values (variadic)
         */
        public function warning( $fmt )
        {
            $args = func_get_args( );
            $this->write( 'W', $args, true );
        }

        /**
         * Issue error level messages (also goes to messages.log)
         *
         * @param string $fmt Format
         * @param mixed ... Format values (variadic)
         */
        public function error( $fmt )
        {
            $args = func_get_args( );
            $this->write( 'E', $args, true );
        }

    }
